import weakref

import gc
from kivy.uix.screenmanager import WipeTransition, FadeTransition

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from mpfmc.transitions.move_in import MoveInTransition
from mpf.tests.MpfTestCase import MpfTestCase

import mpfmc.core
import os


class TestSlidePlayer(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/slide_player'

    def get_config_file(self):
        return 'test_slide_player.yaml'

    def test_slide_with_vars(self):
        self.mc.events.post('show_slide_with_var')
        self.advance_time()
        self.assertEqual("SLIDE WITH VAR asd", self.mc.targets["default"].current_slide.children[0].children[0].text)

    def test_slide_on_default_display(self):
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # now replace that slide at the same priority and make sure it works
        self.mc.events.post('show_slide_4')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_4')

    def test_slide_on_default_display_hardcoded(self):
        self.mc.events.post('show_slide_2')
        self.advance_time()
        self.assertEqual(self.mc.displays['display1'].current_slide_name,
                         'machine_slide_2')

    def test_animation(self):
        self.mc.events.post("show_slide_with_animations")
        self.advance_time()
        self.assertEqual(self.mc.displays['display1'].current_slide_name,
                         'my_slide')
        slide = weakref.ref(self.mc.targets['display1'].current_slide)
        self.assertTrue(slide())

        self.mc.events.post("remove_slide_with_animations")
        self.advance_time()
        self.assertEqual(self.mc.displays['display1'].current_slide_name,
                         'display1_blank')

        self.mc.events.post('show_slide_1')
        self.advance_time()

        gc.collect()
        self.advance_time()
        self.assertFalse(slide())

    def test_slide_on_second_display(self):
        self.mc.events.post('show_slide_3')
        self.advance_time()
        self.assertEqual(self.mc.displays['display2'].current_slide_name,
                         'machine_slide_3')

    def test_priority_from_slide_player(self):
        self.mc.events.post('show_slide_4_p200')
        self.advance_time()
        self.assertEqual(self.mc.displays['display1'].current_slide_name,
                         'machine_slide_4')
        self.assertEqual(self.mc.displays['display1'].current_slide.priority,
                         200)

    def test_force_slide(self):
        self.mc.events.post('show_slide_4_p200')
        self.advance_time()
        self.assertEqual(self.mc.displays['display1'].current_slide_name,
                         'machine_slide_4')
        self.assertEqual(self.mc.displays['display1'].current_slide.priority,
                         200)

        self.mc.events.post('show_slide_1_force')
        self.advance_time()
        self.assertEqual(self.mc.displays['display1'].current_slide_name,
                         'machine_slide_1')
        self.assertEqual(self.mc.displays['display1'].current_slide.priority,
                         0)

    def test_dont_show_slide(self):
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')
        self.assertEqual(self.mc.displays['display1'].current_slide.priority,
                         0)

        # request a higher priority slide, but don't show it
        self.mc.events.post('show_slide_5_dont_show')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')
        self.assertEqual(self.mc.displays['display1'].current_slide.priority,
                         0)

    def test_mode_slide_player(self):
        # set a baseline slide
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # post the slide_player event from the mode. Should not show the slide
        # since the mode is not running
        self.mc.events.post('show_mode1_slide')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # start the mode and then post that event again. The slide should
        # switch
        self.mc.modes['mode1'].start()
        self.mc.events.post('show_mode1_slide')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'mode1_slide')
        slide = weakref.ref(self.mc.targets['display1'].current_slide)
        self.assertTrue(slide())
        # stop the mode and make sure the slide is removed
        num_slides = len(self.mc.targets['display1'].slides)
        self.mc.modes['mode1'].stop()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')
        self.assertEqual(len(self.mc.targets['display1'].slides),
                         num_slides - 1)

        self.advance_time(.1)

        gc.collect()
        self.assertFalse(slide())

        # post the slide_player event from the mode. Should not show the slide
        # since the mode is not running
        self.mc.events.post('show_mode1_slide')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # show a priority 200 slide from the machine config
        self.mc.events.post('show_slide_4_p200')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_4')
        self.assertEqual(self.mc.targets['display1'].current_slide.priority,
                         200)

        # start the mode again (priority 500)
        self.mc.modes['mode1'].start()

        # show a slide, but priority 150 which means the slide will not be
        # shown
        self.mc.events.post('show_mode1_slide_2')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_4')
        self.assertEqual(self.mc.targets['display1'].current_slide.priority,
                         200)

        # now kill the current slide and the mode slide should show
        self.mc.targets['display1'].remove_slide('machine_slide_4')
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'mode1_slide_2')
        self.assertEqual(self.mc.targets['display1'].current_slide.priority,
                         150)

    def test_slides_created_in_slide_player(self):
        # Anon slides are where the widgets are listed in the slide_player
        # section of a config file or the slides section of a show

        self.mc.events.post('anon_slide_dict')
        self.advance_time()

        self.mc.events.post('anon_slide_list')
        self.advance_time()

        self.mc.events.post('anon_slide_widgets')
        self.advance_time()

        slide = weakref.ref(self.mc.targets['display1'].current_slide)
        self.assertTrue(slide())

        self.mc.events.post('anon_slide_widgets2')
        self.advance_time(1)
        slide2 = weakref.ref(self.mc.targets['display1'].current_slide)

        gc.collect()

        self.assertFalse(slide())
        self.assertTrue(slide2())

    def test_expire_in_slide(self):
        # tests that slide expire time works when configured in a slide
        self.mc.events.post('base_slide_no_expire')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_6')

        self.mc.events.post('show_slide_7')  # expire 1s
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_7')

        self.advance_time(1)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_6')

    def test_expire_in_slide_player(self):
        # tests that expire time works when configured in the slide player
        self.mc.events.post('base_slide_no_expire')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_6')

        self.mc.events.post('new_slide_expire')  # expire 1s
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        self.advance_time(1)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_6')

    def test_expire_with_transition_out_in_slide(self):
        # Tests a slide expiring where the expiring slide has a transition
        self.mc.events.post('base_slide_no_expire')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_6')

        # show a slide which expires in 1 sec, and has a transition out set
        self.mc.events.post('show_slide_8')
        self.advance_time(.1)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_8')

        # advance to after this slide_8 expired, transition should be in effect
        self.advance_time(1)
        self.assertTrue(isinstance(self.mc.targets['display1'].transition,
                                   WipeTransition))

        # advance to transition done, should be back to the original slide
        self.advance_time(1)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_6')

    def test_current_slide_transition_out(self):
        # Tests a new slide with no transition, but the current slide has one,
        # so it uses that

        # show a slide, no expire, but with transition out
        self.mc.events.post('show_slide_9')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_9')

        # show a new slide with no transition
        self.assertIsNone(self.mc.slides['machine_slide_6']['transition'])
        self.mc.events.post('machine_slide_6')
        self.advance_time()

        # transition from first slide should be happening
        self.assertTrue(isinstance(self.mc.targets['display1'].transition,
                                   MoveInTransition))

    def test_both_slides_transitions(self):
        # current slide has transition out, and new slide has transition, so
        # transition of new slide takes precendence

        # show a slide, no expire, but with transition out
        self.assertEqual(
            self.mc.slides['machine_slide_8']['transition_out']['type'],
            'wipe')
        self.mc.events.post('show_slide_8')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_8')

        # show a new slide with a different transition in
        self.assertEqual(
            self.mc.slides['machine_slide_9']['transition']['type'], 'move_in')
        self.mc.events.post('show_slide_9')
        self.advance_time()

        # transition from second slide should be happening
        self.assertTrue(isinstance(self.mc.targets['display1'].transition,
                                   MoveInTransition))

    def test_transition_in_slide_player(self):
        # transition is specified in slide player for slide that does not have
        # transition

        # show a base slide with no transition
        self.assertIsNone(self.mc.slides['machine_slide_4']['transition'])
        self.mc.events.post('machine_slide_4')
        self.advance_time()

        # show a second slide where the slide has no transition, but the
        # slide player does have a transition
        self.assertIsNone(self.mc.slides['machine_slide_5']['transition'])
        self.mc.events.post('show_slide_5_with_transition')
        self.advance_time()

        # make sure the transition is happening
        self.assertTrue(isinstance(self.mc.targets['display1'].transition,
                                   FadeTransition))

    def test_transition_in_slide_player_override(self):
        # transition in slide player for slide that already has a transition.
        # the slide player transition should override the slide one

        # show a base slide with no transition
        self.assertIsNone(self.mc.slides['machine_slide_4']['transition'])
        self.mc.events.post('machine_slide_4')
        self.advance_time()

        # show a second slide where the slide has a transition, but the
        # slide player has a different transition, so the slide player
        # should take precedence
        self.assertEqual(
            self.mc.slides['machine_slide_9']['transition']['type'], 'move_in')
        self.mc.events.post('show_slide_5_with_transition')
        self.advance_time()

        # make sure the transition from the slide player is happening
        self.assertTrue(isinstance(self.mc.targets['display1'].transition,
                                   FadeTransition))

    def test_slide_show(self):
        # tests the 'show' feature of a slide. This is not a slide show, but
        # rather a setting which controls whether a slide is shown right away
        # or not

        # show a base slide
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # post new slide, but with show=False, so it should not show
        self.mc.events.post('slide_2_dont_show')
        self.advance_time()
        # Should still be slide 1
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

    def test_slide_removal(self):
        # Also test slide events
        self.mock_event('slide_machine_slide_1_active')
        self.mock_event('slide_machine_slide_1_created')
        self.mock_event('slide_machine_slide_1_removed')
        self.mock_event('slide_machine_slide_4_active')
        self.mock_event('slide_machine_slide_4_created')
        self.mock_event('slide_machine_slide_4_removed')

        # show a base slide
        self.mc.events.post('show_slide_1')
        self.advance_time(0.3)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        self.assertEventCalled('slide_machine_slide_1_created')
        self.assertEventCalled('slide_machine_slide_1_active')
        self.assertEventNotCalled('slide_machine_slide_1_removed')
        self.assertEventNotCalled('slide_machine_slide_4_created')
        self.assertEventNotCalled('slide_machine_slide_4_active')
        self.assertEventNotCalled('slide_machine_slide_4_removed')

        # show another slide
        self.mc.events.post('show_slide_4')
        self.advance_time(0.3)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_4')

        self.assertEventCalled('slide_machine_slide_1_created', 1)
        self.assertEventCalled('slide_machine_slide_1_active', 1)
        self.assertEventNotCalled('slide_machine_slide_1_removed')
        self.assertEventCalled('slide_machine_slide_4_created', 1)
        self.assertEventCalled('slide_machine_slide_4_active', 1)
        self.assertEventNotCalled('slide_machine_slide_4_removed')

        # make sure base slide comes back
        self.mc.events.post('remove_slide_4')
        self.advance_time(0.3)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        self.assertEventCalled('slide_machine_slide_1_created', 1)
        self.assertEventCalled('slide_machine_slide_1_active', 2)
        self.assertEventNotCalled('slide_machine_slide_1_removed')
        self.assertEventCalled('slide_machine_slide_4_created', 1)
        self.assertEventCalled('slide_machine_slide_4_active', 1)
        self.assertEventCalled('slide_machine_slide_4_removed', 1)

    def test_slide_removal_new_transition(self):
        # show a base slide
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # show a slide with not transition out
        self.assertIsNone(self.mc.slides['machine_slide_4']['transition_out'])
        self.mc.events.post('show_slide_4')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_4')

        # remove that slide with a transition
        self.mc.events.post('remove_slide_4_with_transition')
        self.advance_time(.1)

        # make sure the transition is taking effect
        self.assertTrue(isinstance(self.mc.targets['display1'].transition,
                                   WipeTransition))

        # original slide is back
        self.advance_time(1)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

    def test_slide_removal_existing_transition(self):
        # show a base slide
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # show a slide which has a transition out
        self.assertEqual(
            self.mc.slides['machine_slide_8']['transition_out']['type'],
            'wipe')
        self.mc.events.post('show_slide_8')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_8')

        # post an event which does not have a transition
        self.mc.events.post('remove_slide_8')
        self.advance_time(.1)

        # make sure the transition is taking effect
        self.assertTrue(isinstance(self.mc.targets['display1'].transition,
                                   WipeTransition))

        # original slide is back
        self.advance_time(1)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

    def test_slide_removal_override_transition(self):
        # show a base slide
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # show a slide which has a wipe transition
        self.assertEqual(
            self.mc.slides['machine_slide_8']['transition_out']['type'],
            'wipe')
        self.mc.events.post('show_slide_8')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_8')

        # remove slide with a fade transition
        self.mc.events.post('remove_slide_8_fade')
        self.advance_time(.1)

        # make sure it uses the fade transition from the slide player
        self.assertTrue(isinstance(self.mc.targets['display1'].transition,
                                   FadeTransition))

        # original slide should be back
        self.advance_time(1)
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

    def test_removing_last_slide(self):
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'machine_slide_1')
        self.advance_time()
        self.mc.targets['default'].remove_slide('machine_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'display1_blank')
        self.assertEqual(1, len(self.mc.targets['default'].screens))

    def test_expire_non_current_slide(self):
        self.mc.events.post('slide1_expire_1s')
        self.advance_time(.1)
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'machine_slide_1')

        # show slide 2 which should expire in 1s
        self.mc.events.post('slide2_expire_1s')
        self.advance_time(.1)
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'machine_slide_2')

        self.advance_time(1)

        # should be back to blank, because slide1 expired while slide 2 was up
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'display1_blank')
        self.assertEqual(1, len(self.mc.targets['default'].screens))

    def test_remove_already_removed_slide(self):
        self.mc.events.post('slide1_expire_1s')
        self.advance_time(.1)
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'machine_slide_1')

        # grab a reference to this slide
        slide1 = self.mc.targets['default'].current_slide

        self.advance_time(1)

        # should be blank, because slide1 expired
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'display1_blank')
        self.assertEqual(1, len(self.mc.targets['default'].screens))

        # now try to call the now-gone slide's remove

        slide1.remove()
        self.advance_time()

    def test_animation_triggers(self):
        bcp_command1 = ('register_trigger', None, {'event': 'flash_widget_1'})
        bcp_command2 = ('register_trigger', None, {'event': 'flash_widget_2'})
        self.assertNotIn(bcp_command1, self.sent_bcp_commands)
        self.assertNotIn(bcp_command2, self.sent_bcp_commands)

        self.mc.events.post("client_connected")
        self.advance_time()
        self.assertIn(bcp_command1, self.sent_bcp_commands)
        self.assertIn(bcp_command2, self.sent_bcp_commands)

    def test_play_multiple_times(self):
        # set a baseline slide
        self.mc.events.post('show_slide_1')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'machine_slide_1')

        # start the mode and then post that event again. The slide should
        # switch
        self.mc.modes['mode1'].start()

        self.mc.events.post('show_mode1_slide')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name, 'mode1_slide')
        slide = weakref.ref(self.mc.targets['display1'].current_slide)

        for i in range(10):
            self.mc.events.post('show_mode1_slide')
            self.advance_time()
            self.assertEqual(self.mc.targets['display1'].current_slide_name, 'mode1_slide')
            # run garbage collector
            gc.collect()
            # weak ref to the slide should be none
            self.assertIsNone(slide())
            # build weak ref to curent slide
            slide = weakref.ref(self.mc.targets['display1'].current_slide)

    def test_conditional_slide(self):
        self.mc.events.post('show_conditional_slide', var='foo')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'slide_condition_foo')
        self.mc.events.post('show_conditional_slide', var='bar')
        self.advance_time()
        self.assertEqual(self.mc.targets['display1'].current_slide_name,
                         'slide_condition_bar')


class TestMpfSlidePlayer(MpfTestCase):

    # runs the MPF tests (and not the MPF-MC ones) to test the MPF side of the
    # slide player plugin

    def __init__(self, methodName):
        super().__init__(methodName)
        # remove config patch which disables bcp
        del self.machine_config_patches['bcp']

    def get_absolute_machine_path(self):
        # override the base to we set the patch based on the mpfmc location
        return os.path.abspath(os.path.join(
            mpfmc.core.__path__[0], os.pardir, self.get_machine_path()))

    def get_enable_plugins(self):
        return True

    def get_config_file(self):
        return 'test_slide_player.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/slide_player/'

    # todo add tests



from typing import Optional

from kivy.uix.effectwidget import EffectWidget

from kivy.uix.relativelayout import RelativeLayout

from mpfmc.uix.widget import Widget
from mpfmc.uix.display import DisplayOutput

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc
    from mpfmc.uix.slide import Slide


class DisplayWidget(Widget, RelativeLayout):

    """A widget showing a display."""

    widget_type_name = 'Display'
    animation_properties = ('x', 'y', 'pos')

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        del kwargs
        self.display = mc.displays[config['source_display']]

        super().__init__(mc=mc, config=config, key=key)

        if 'effects' in self.config and self.config['effects']:
            self.effects = EffectWidget(pos=self.pos, size_hint=(1, 1))
            self.effects.key = None
            self._add_effects(self.config['effects'])
        else:
            self.effects = RelativeLayout(pos=self.pos, size_hint=(1, 1))

        # Establish link between display and this display widget
        self.display_output = DisplayOutput(self.effects, self.display)
        self.effects.add_widget(self.display_output)
        self.add_widget(self.effects)
        self.display_output.add_display_source(self.display)

    def prepare_for_removal(self):
        """Remove display and effects."""
        super().prepare_for_removal()
        self.remove_widget(self.effects)
        self.display_output.remove_display_source(self.display)

    def __repr__(self) -> str:  # pragma: no cover
        try:
            return '<DisplayWidget size={}, pos={}, source={}>'.format(
                self.size, self.pos, self.display.name)
        except AttributeError:
            return '<DisplayWidget size={}, source=(none)>'.format(self.size)

    def on_pos(self, instance, pos):
        del instance
        self.effects.pos = pos

    def _add_effects(self, config: Optional[list]) -> None:
        """Adds any effects specified in the config for this display widget."""
        if config:
            effects_list = list()
            for effect_config in config:
                effect_config['width'] = self.width
                effect_config['height'] = self.height
                effects_list.extend(self.mc.effects_manager.get_effect(effect_config))

            self.effects.effects = effects_list

    def get_display(self):
        """List display."""
        return self.display

    @property
    def current_slide(self) -> Optional["Slide"]:
        """The current slide shown on the linked display."""
        if self.display:
            return self.display.current_slide
        else:
            return None

    @property
    def current_slide_name(self) -> Optional[str]:
        """The name of the current slide shown on the linked display."""
        if self.display:
            return self.display.current_slide_name
        else:
            return None


widget_classes = [DisplayWidget]



from mpf.config_players.plugin_player import PluginPlayer


class MpfSoundPlayer(PluginPlayer):
    """Base class for part of the sound player which runs as part of MPF.

    Note: This class is loaded by MPF and everything in it is in the context of
    MPF, not the mpf-mc. MPF finds this instance because the mpf-mc setup.py
    has the following entry_point configured:

        sound_player=mpfmc.config_players.sound_player:register_with_mpf

    """
    config_file_section = 'sound_player'
    show_section = 'sounds'

    def get_express_config(self, value):
        """Parse express config."""
        return {"action": value}


player_cls = MpfSoundPlayer


def register_with_mpf(machine):
    """Registers the sound player plug-in with MPF"""
    return 'sound', MpfSoundPlayer(machine)



from kivy.core.window import Window as KivyWindow
from mpfmc.core.keyboard import Keyboard
from mpfmc.uix.display import Display
from mpfmc.widgets.display import DisplayWidget

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


class Window:

    @staticmethod
    def set_source_display(display: "DisplayWidget") -> None:
        """Set the source display for the main window."""

        # Clean up any existing objects in the window
        KivyWindow.clear()

        for widget in KivyWindow.children:
            KivyWindow.remove_widget(widget)

        # Add the new display to the window
        KivyWindow.add_widget(display)

        # Make sure the display is re-sized whenever the window changes size
        KivyWindow.bind(system_size=Window.on_size)

    @staticmethod
    def initialize(mc: "MpfMc") -> None:
        mc.icon = mc.machine_config['window']['icon']
        mc.title = mc.machine_config['window']['title']

        # Set the source based on the window: source_display: setting.
        # If that's not valid, and there's a display called "window", use it
        # Otherwise use the default.
        try:
            display = mc.displays[mc.machine_config['window']['source_display']]
        except KeyError:
            display = mc.targets['default']

        # We need the window to map to a Display instance, so no matter what
        # we're passed, we keep on moving up until we find the actual display.
        while not isinstance(display, Display):
            display = display.parent

        config = {
            'type': DisplayWidget.widget_type_name.lower(),
            'source_display': display.name,
            'width': KivyWindow.width,
            'height': KivyWindow.height,
        }

        config = mc.config_validator.validate_config('widgets:{}'.format(
            DisplayWidget.widget_type_name.lower()), config, base_spec='widgets:common')

        if 'effects' in mc.machine_config['window']:
            config['effects'] = mc.effects_manager.validate_effects(mc.machine_config['window']['effects'])

        display_widget = DisplayWidget(mc, config=config)
        display_widget.parent.remove_widget(display_widget)
        Window.set_source_display(display_widget)

        if (display.width / display.height !=
                KivyWindow.width / KivyWindow.height):
            mc.log.warning(
                "ASPECT RATIO MISMATCH: The on-screen window is not the same "
                "aspect ratio as the display called '{}'. The logical display "
                "will be scaled from {}x{} to the window size {}x{}, and a "
                "black bar will make up the difference".format(
                    display.name, display.width, display.height,
                    KivyWindow.width, KivyWindow.height))

        if 'keyboard' in mc.machine_config:
            mc.keyboard = Keyboard(mc)

    @staticmethod
    def on_size(instance, value):
        del instance

        for widget in KivyWindow.children:
            widget.size = value



from typing import Union
from mpfmc.core.config_collection import ConfigCollection


class SlideCollection(ConfigCollection):

    config_section = 'slides'
    collection = 'slides'
    class_label = 'SlideConfig'

    def process_config(self, config: Union[dict, list]) -> dict:
        # config is localized to an single slide name entry

        if isinstance(config, list):
            new_dict = dict()
            new_dict['widgets'] = config
            config = new_dict

        elif isinstance(config, dict):
            if 'widgets' not in config:
                new_dict = dict()
                new_dict['widgets'] = [config]
                config = new_dict

            elif not isinstance(config['widgets'], list):
                if config['widgets']:
                    config['widgets'] = [config['widgets']]
                else:
                    config['widgets'] = []

        for i, widget in enumerate(config['widgets']):
            # since dict is mutable it updates in place
            config['widgets'][i] = self.mc.widgets.process_widget(widget)

        config = self.mc.config_validator.validate_config('slides', config)
        config = self.mc.transition_manager.validate_transitions(config)

        return config

    def validate_config(self, config):
        # since dict is mutable it updates in place
        self.mc.widgets.validate_config(config['widgets'])


collection_cls = SlideCollection



"""Widget player which can add and remove widgets from slides."""

from copy import deepcopy
from mpf.core.events import EventHandlerKey
from mpfmc.core.mc_config_player import McConfigPlayer
from mpfmc.uix.widget import create_widget_objects_from_library


class SlideNotActiveError(Exception):

    """Slide is not currently active but exists."""

    def __init__(self, slide_name, *args, **kwargs):
        self.slide_name = slide_name
        super().__init__(*args, **kwargs)


class McWidgetPlayer(McConfigPlayer):

    """Base class for the Widget Player that runs on the mpf-mc side of things.

    It receives all of its instructions via BCP from an MpfWidgetPlayer
    instance running as part of MPF.
    """

    config_file_section = 'widget_player'
    show_section = 'widgets'
    machine_collection_name = None  # widgets are not real devices currently

    def _get_slide(self, s):
        slide = None
        if s.get('slide'):
            slide_name = s['slide']
            try:
                slide = self.machine.active_slides[slide_name]
            except KeyError:
                # check if slide does exist
                if slide_name not in self.machine.slides:
                    raise KeyError(
                        "Widget Player Error: Slide name '{}' is not valid "
                        "slide. Widget config that caused this: "
                        "{}".format(slide_name, s))

                raise SlideNotActiveError(slide_name=slide_name)

        return slide

    # pylint: disable-msg=too-many-arguments
    def _action_add(self, s, instance_dict, widget, context, play_kwargs):
        if not s['key']:
            try:
                s['key'] = s['widget_settings'].pop('key')
            except (KeyError, AttributeError):
                s['key'] = context + "-" + widget

        if s.get('target'):

            try:
                target = self.machine.targets[s['target']]

            except KeyError:
                raise KeyError("Widget player invalid target '{}'. Current "
                               "valid targets: {} Settings: "
                               "{} Widget: {} Context: {} Play kwargs: {}".
                               format(self.machine.targets, s['target'], s,
                                      widget, context, play_kwargs))

            # remove any instances of this widget from this slide first
            target.remove_widgets_by_key(s['key'])

            target.add_widgets_to_current_slide(
                create_widget_objects_from_library(mc=self.machine,
                                                   name=widget,
                                                   play_kwargs=play_kwargs,
                                                   **s))  # todo
            if not s['key'] in instance_dict:
                instance_dict[s['key']] = True

            return

        priority = s.pop("priority", 1)

        if 'slide' in s and s['slide']:
            if s['key'] in instance_dict and isinstance(instance_dict[s['key']], EventHandlerKey):
                self.machine.events.remove_handler_by_key(instance_dict[s['key']])
            handler = self.machine.events.add_handler(
                "slide_{}_created".format(s['slide']), self._add_widget_to_slide_when_active,
                slide_name=s['slide'], widget=widget, s=s, priority=priority, play_kwargs=play_kwargs)
            instance_dict[s['key']] = handler

        try:
            slide = self._get_slide(s)
        except SlideNotActiveError:
            return

        if not slide:
            slide = self.machine.targets['default'].current_slide

        if not slide:  # pragma: no cover
            raise ValueError("Cannot add widget. No current slide")

        # remove from any slides since we are not targeting a specific slide
        self._remove_widget_by_key(s['key'])
        # add widget
        slide.add_widgets_from_library(name=widget, play_kwargs=play_kwargs, **s)
        if not s['key'] in instance_dict:
            instance_dict[s['key']] = True

    def _validate_config_item(self, device, device_settings):
        validated_dict = super()._validate_config_item(device, device_settings)
        for widget, widget_settings in validated_dict.items():
            if widget_settings['action'] == "add" and widget not in self.machine.widgets:
                raise AssertionError("Unknown widget {}".format(widget))
            if widget_settings["target"] and widget_settings["target"] not in self.machine.targets:
                raise AssertionError("Unknown target {}".format(widget_settings["target"]))

        return validated_dict

    def _action_remove(self, s, instance_dict, widget, context):
        if s['key']:
            key = s['key']
        else:
            key = context + "-" + widget

        if key in instance_dict and isinstance(instance_dict[key], EventHandlerKey):
            self.machine.events.remove_handler_by_key(instance_dict[key])

        if key in instance_dict:
            del instance_dict[key]

        try:
            slide = self._get_slide(s)
        except SlideNotActiveError:
            return

        if slide:
            slide.remove_widgets_by_key(key)
        else:
            self._remove_widget_by_key(key)

    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Play widgets."""
        # **kwargs since this is an event callback
        del priority
        del calling_context
        settings = deepcopy(settings)
        instance_dict = self._get_instance_dict(context)

        if 'widgets' in settings:
            settings = settings['widgets']

        for widget, s in settings.items():
            action = s.pop('action')
            assert action in ('add', 'remove', 'update')

            if action == "add":
                self._action_add(s, instance_dict, widget, context, kwargs)
            elif action == "remove":
                self._action_remove(s, instance_dict, widget, context)
            elif action == "update":
                self._action_remove(s, instance_dict, widget, context)
                self._action_add(s, instance_dict, widget, context, kwargs)
            else:
                raise AssertionError("Invalid action {} in {}".format(action, widget))

    def _remove_widget_by_key(self, key):
        """Remove widget by key."""
        for target in self.machine.targets.values():
            target.remove_widgets_by_key(key)

    def _add_widget_to_slide_when_active(self, slide_name, widget, s, play_kwargs, **kwargs):
        del kwargs
        if slide_name in self.machine.active_slides:
            slide = self.machine.active_slides[slide_name]
            # remove any instances of this widget from this slide first
            slide.remove_widgets_by_key(s['key'])
            # add widget
            slide.add_widgets_from_library(name=widget, play_kwargs=play_kwargs, **s)

    def get_express_config(self, value):
        """Parse express config."""
        return dict(widget=value)

    def clear_context(self, context):
        """Clear context."""
        instance_dict = self._get_instance_dict(context)
        for key in instance_dict:
            if isinstance(instance_dict[key], EventHandlerKey):
                self.machine.events.remove_handler_by_key(instance_dict[key])

            self._remove_widget_by_key(key)

        self._reset_instance_dict(context)


mc_player_cls = McWidgetPlayer



"""Contains the sound loop config player class"""

from copy import deepcopy
from mpfmc.core.mc_config_player import McConfigPlayer


class McSoundLoopPlayer(McConfigPlayer):
    """Base class for the Sound Loop Player that runs on the mpf-mc side of things.
    It receives all of its instructions via BCP from a MpfSoundLoopPlayer instance
    running as part of MPF.

    The sound_loop_player: section of a config file (either the machine-wide or
    a mode-specific config) looks like this:

    sound_loop_player:
        <event_name>:
            <track_name>:
                <sound_loop_set_settings>: ...

    Express config is not supported in the sound_loop_player.

    If you want to control other settings (such as track, priority, volume,
    loops, etc.), enter the track name on the next line and the settings
    indented under it, like this:

    sound_loop_player:
        some_event:
            track_name:
                volume: 0.35
                action: play_layer
                layer: 2

Here are several various examples:

    sound_loop_player:
        some_event:
            loops:
                sound_loop_set: basic_beat
                volume: 0.65

        some_event2:
            loops:
                action: set_volume
                volume: -4.5 db

    """
    config_file_section = 'sound_loop_player'
    show_section = 'sound_loop_sets'
    machine_collection_name = 'sound_loop_sets'

    # pylint: disable=invalid-name
    def play(self, settings, context, calling_context, priority=0, **kwargs):
        """Plays a validated section from a sound_loop_player: section of a
        config file or the sound_loops: section of a show.

        The config must be validated.
        """
        del calling_context
        settings = deepcopy(settings)

        self.machine.log.info("SoundLoopPlayer: Play called with settings=%s", settings)

        for track_name, player_settings in settings.items():
            player_settings.update(kwargs)

            track = self.machine.sound_system.audio_interface.get_track_by_name(track_name)

            # Determine action to perform
            if player_settings['action'].lower() == 'play':
                if player_settings['sound_loop_set'] not in self.machine.sound_loop_sets:
                    self.machine.log.error("SoundLoopPlayer: The specified sound loop set "
                                           "does not exist ('{}').".format(player_settings['sound_loop_set']))
                    return
                try:
                    loop_set = self.machine.sound_loop_sets[player_settings['sound_loop_set']]
                    track.play_sound_loop_set(loop_set, context, player_settings)
                except Exception as ex:
                    raise Exception(ex)

            elif player_settings['action'].lower() == 'stop':
                player_settings.setdefault('fade_out', None)
                track.stop_current_sound_loop_set(player_settings['fade_out'])

            elif player_settings['action'].lower() == 'stop_looping':
                track.stop_looping_current_sound_loop_set()

            elif player_settings['action'].lower() == 'jump_to':
                player_settings.setdefault('time', 0)
                track.jump_to_time_current_sound_loop_set(player_settings['time'])

            elif player_settings['action'].lower() == 'set_volume':
                # TODO: Implement setting current loop_set volume
                pass

            elif player_settings['action'].lower() == 'play_layer':
                player_settings.setdefault('volume', None)
                track.play_layer(player_settings['layer'], player_settings['fade_in'], player_settings['timing'],
                                 volume=player_settings['volume'])

            elif player_settings['action'].lower() == 'stop_layer':
                track.stop_layer(player_settings['layer'], player_settings['fade_out'])

            elif player_settings['action'].lower() == 'stop_looping_layer':
                track.stop_looping_layer(player_settings['layer'])

            elif player_settings['action'].lower() == 'set_layer_volume':
                # TODO: Implement setting layer volume
                pass

            else:
                self.machine.log.error("SoundLoopPlayer: The specified action "
                                       "is not valid ('{}').".format(player_settings['action']))

    def get_express_config(self, value):
        """There is no express config for the sound loop player."""
        self.machine.log.error("SoundLoopPlayer: Express config is not supported: "
                               "'{}'".format(value))
        raise AssertionError("Sound Loop Player does not support express config")

    # pylint: disable=too-many-branches
    def validate_config(self, config):
        """Validates the sound_loop_player: section of a config file (either a
        machine-wide config or a mode config).

        Args:
            config: A dict of the contents of the sound_loop_player section
            from the config file. It's assumed that keys are event names, and
            values are settings for what the sound_loop_player should do when
            that event is posted.

        Returns: A dict a validated entries.

        This method overrides the base method since the sound_loop_player has
        unique options.

        """
        validated_config = dict()

        # No need to validate if sound system is not enabled, just return empty dict
        if self.machine.sound_system is None or self.machine.sound_system.audio_interface is None:
            return validated_config

        for event, settings in config.items():
            validated_config[event] = dict()

            for track_name, player_settings in settings.items():

                # Validate the specified track name is a sound_loop track
                if self.machine.sound_system.audio_interface.get_track_type(track_name) != "sound_loop":
                    raise ValueError("SoundLoopPlayer: An invalid audio track '{}' is specified for event '{}' "
                                     "(only sound_loop audio tracks are supported).".format(track_name, event))

                validated_config[event].update(self._validate_config_item(track_name, player_settings))

        return validated_config

    def _validate_config_item(self, device, device_settings):
        """Validates the config when in a show or in a player"""

        # event contains the event name that triggers the sound_loop_player action
        # Validate the settings against the config spec

        # First validate the action item (since it will be used to validate the rest
        # of the config)
        if 'action' in device_settings:
            device_settings['action'] = self.machine.config_validator.validate_config_item(
                self.machine.config_validator.config_spec['sound_loop_player']['action'],
                'sound_loop_player:{}'.format(device),
                device_settings['action'])
        else:
            device_settings['action'] = self.machine.config_validator.validate_config_item(
                self.machine.config_validator.config_spec['sound_loop_player']['action'],
                'sound_loop_player:{}'.format(device))

        validated_settings = self.machine.config_validator.validate_config(
            'sound_loop_player_actions:{}'.format(device_settings['action']).lower(),
            device_settings)

        # Remove any items from the settings that were not explicitly provided in the
        # sound_loop_player config section (only want to override sound settings explicitly
        # and not with any default values).  The default values for these items are not
        # legal values and therefore we know the user did not provide them.
        if 'volume' in validated_settings and validated_settings['volume'] is None:
            del validated_settings['volume']
        if 'fade_in' in validated_settings and validated_settings['fade_in'] is None:
            del validated_settings['fade_in']
        if 'fade_out' in validated_settings and validated_settings['fade_out'] is None:
            del validated_settings['fade_out']
        if 'events_when_played' in validated_settings and len(validated_settings['events_when_played']) == 1 and \
                validated_settings['events_when_played'][0] == 'use_sound_loop_setting':
            del validated_settings['events_when_played']
        if 'events_when_stopped' in validated_settings and len(validated_settings['events_when_stopped']) == 1 and \
                validated_settings['events_when_stopped'][0] == 'use_sound_loop_setting':
            del validated_settings['events_when_stopped']
        if 'events_when_looping' in validated_settings and len(validated_settings['events_when_looping']) == 1 and \
                validated_settings['events_when_looping'][0] == 'use_sound_loop_setting':
            del validated_settings['events_when_looping']
        if 'mode_end_action' in validated_settings and (validated_settings['mode_end_action'] is None or
                                                        validated_settings['mode_end_action'] ==
                                                        'use_sound_loop_setting'):
            del validated_settings['mode_end_action']

        validated_dict = dict()
        validated_dict[device] = validated_settings
        return validated_dict

    def clear_context(self, context):
        """Stop all sounds from this context."""
        instance_dict = self._get_instance_dict(context)
        # Iterate over a copy of the dictionary values since it may be modified
        # during the iteration process.
        self.machine.log.debug("SoundLoopPlayer: Clearing context - applying mode_end_action for active sound loop set")
        # TODO: Determine proper action to take while clearing context
        for sound_instance in list(instance_dict.values()):
            if sound_instance.stop_on_mode_end:
                sound_instance.stop()
            else:
                sound_instance.stop_looping()

        self._reset_instance_dict(context)

    def on_sound_instance_finished(self, sound_instance_id, context, **kwargs):
        """Callback function that is called when a sound instance triggered by the sound_player
        is finished. Remove the specified sound instance from the list of current instances
        started by the sound_player."""
        del kwargs
        instance_dict = self._get_instance_dict(context)
        if sound_instance_id in instance_dict:
            del instance_dict[sound_instance_id]


mc_player_cls = McSoundLoopPlayer



from mpfmc.core.utils import percent_to_float
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestUtils(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/utils'

    def get_config_file(self):
        return 'test_utils.yaml'

    def test_percent_to_float(self):
        num = 1
        total = 1
        self.assertEqual(percent_to_float(num, total), 1.0)

        num = 1
        total = 2
        self.assertEqual(percent_to_float(num, total), 1.0)

        num = 0
        total = 2
        self.assertEqual(percent_to_float(num, total), 0.0)

        num = '1'
        total = 1
        self.assertEqual(percent_to_float(num, total), 1.0)

        num = '1'
        total = 2
        self.assertEqual(percent_to_float(num, total), 1.0)

        num = '0'
        total = 2
        self.assertEqual(percent_to_float(num, total), 0.0)

        num = '100%'
        total = 1
        self.assertEqual(percent_to_float(num, total), 1.0)

        num = '100%'
        total = 2
        self.assertEqual(percent_to_float(num, total), 2.0)

        num = '0%'
        total = 2
        self.assertEqual(percent_to_float(num, total), 0.0)

        num = '25%'
        total = 800
        self.assertEqual(percent_to_float(num, total), 200.0)

        num = '200%'
        total = 1
        self.assertEqual(percent_to_float(num, total), 2.0)



from kivy.core.text import LabelBase

from mpfmc.assets.bitmap_font import BitmapFontAsset
from mpfmc.uix.bitmap_font.bitmap_font import _SurfaceContainer
from mpfmc.core.mc import MpfMc


class LabelBitmapFont(LabelBase):

    def __init__(self, mc: MpfMc, font_name, text='', font_kerning=True, **kwargs):
        del kwargs
        self.mc = mc
        self._surface = None
        super().__init__(text=text, font_name=font_name, font_kerning=font_kerning)

    def get_font_asset(self) -> BitmapFontAsset:
        """Return the bitmap font asset used for this label."""
        return self.mc.bitmap_fonts[self.options['font_name_r']]

    def resolve_font_name(self):
        """Resolves the supplied bitmap font name."""
        options = self.options
        fontname = options['font_name']

        if fontname in self.mc.bitmap_fonts:
            options['font_name_r'] = fontname
        else:
            raise ValueError('LabelBitmapFont: font_name %s not found in bitmap_fonts.' % fontname)

    def get_extents(self, text):
        options = self.options
        bitmap_font = self.get_font_asset()
        return bitmap_font.get_extents(text, options['font_kerning'])

    def get_descent(self):
        bitmap_font = self.get_font_asset()
        return bitmap_font.get_descent()

    def get_ascent(self):
        bitmap_font = self.get_font_asset()
        return bitmap_font.get_ascent()

    def _render_begin(self):
        self._surface = _SurfaceContainer(self.size[0], self.size[1])

    def _render_text(self, text, x, y):
        self._surface.render(self, text, x, y)

    def _render_end(self):
        return self._surface.get_data()



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestDmd(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/dmd'

    def get_config_file(self):
        return 'test_color_dmd.yaml'

    def test_color_dmd(self):
        self.advance_time()
        self.assertIn('dmd', self.mc.targets)

        self.mc.events.post('slide1')
        self.mc.events.post('dmd_slide')
        self.advance_time(2)

    def test_color_and_monochrome_dmd(self):
        self.advance_time()
        self.assertIn('dmd', self.mc.targets)

        self.mc.events.post('slide2')
        self.mc.events.post('dmd_slide')
        self.advance_time(4)



import importlib

from kivy.animation import AnimationTransition
from kivy.properties import StringProperty
from kivy.uix.screenmanager import TransitionBase
from kivy.uix.screenmanager import (WipeTransition, SwapTransition,
                                    FadeTransition, FallOutTransition,
                                    RiseInTransition, CardTransition,
                                    NoTransition)


class TransitionManager:
    def __init__(self, mc):
        self.mc = mc
        self._transitions = dict()
        self._register_mpf_transitions()
        self._register_kivy_transitions()

    @property
    def transitions(self):
        return self._transitions

    def register_transition(self, name, transition_cls):
        self._transitions[name] = transition_cls

    def get_transition(self, transition_config=None):
        if transition_config:
            # The kivy shader transitions can't accept unexpected kwargs
            kwargs = transition_config.copy()
            kwargs.pop('type')
            return self._transitions[transition_config['type']](**kwargs)
        else:
            return NoTransition()

    def _register_mpf_transitions(self):
        for t in self.mc.machine_config['mpf-mc']['mpf_transition_modules']:
            i = importlib.import_module('mpfmc.transitions.{}'.format(t))
            self.register_transition(getattr(i, 'name'),
                                     getattr(i, 'transition_cls'))

    def _register_kivy_transitions(self):
        self.register_transition('wipe', WipeTransition)
        self.register_transition('swap', SwapTransition)
        self.register_transition('fade', FadeTransition)
        self.register_transition('fade_back', FallOutTransition)
        self.register_transition('rise_in', RiseInTransition)
        self.register_transition('card', CardTransition)
        self.register_transition('none', NoTransition)

    def validate_transitions(self, config):

        if 'transition' in config:
            if not isinstance(config['transition'], dict):
                config['transition'] = dict(type=config['transition'])

            try:
                config['transition'] = (
                    self.mc.config_validator.validate_config(
                        'transitions:{}'.format(config['transition']['type']),
                        config['transition']))

            except KeyError:
                raise ValueError('transition: section of config requires a'
                                 ' "type:" setting')
        else:
            config['transition'] = None

        if 'transition_out' in config:
            if not isinstance(config['transition_out'], dict):
                config['transition_out'] = dict(type=config['transition_out'])

            try:
                config['transition_out'] = (
                    self.mc.config_validator.validate_config(
                        'transitions:{}'.format(
                            config['transition_out']['type']),
                        config['transition_out']))

            except KeyError:
                raise ValueError('transition_out: section of config '
                                 'requires a "type:" setting')
        else:
            config['transition_out'] = None

        return config


class MpfTransition(TransitionBase):
    """Base class for slide transitions in MPF. Use this when writing your
    own custom transitions.

    """
    easing = StringProperty('linear')
    """String name of the animation 'easing' function that is used to
    control the shape of the curve of the animation.

    Default is 'linear'.

    """

    def __init__(self, **config):
        # Use ** here instead of dict so this constructor is compatible with
        # the Kivy shader transitions too.

        for k, v in config.items():
            if hasattr(self, k):
                setattr(self, k, v)

        super().__init__()

    def get_vars(self, progression):
        """Convenience method you can call in your own transition's
        on_progress() method to easily get the local vars you need to write
        your own transition.

        Args:
            progression: Float from 0.0 to 1.0 that indicates how far along
            the transition is.

        Returns:
            * Incoming slide object
            * Outgoing slide object
            * Width of the screen
            * Height of the screen
            * Modified progression value (0.0-1.0) which factors in the easing
              setting that has been applied to this transition.

        """
        return (self.screen_in, self.screen_out,
                self.manager.width, self.manager.height,
                getattr(AnimationTransition, self.easing)(progression))

    def on_complete(self):
        # reset the screen back to its original position
        self.screen_in.pos = self.manager.pos
        self.screen_out.pos = self.manager.pos
        super().on_complete()

        # todo test super().on_complete(). It removes the screen, but is
        # that what we want?

    def on_progress(self, progression):
        raise NotImplementedError



"""Magic events for widgets."""
magic_events = ('add_to_slide',
                'remove_from_slide',
                'pre_show_slide',
                'show_slide',
                'pre_slide_leave',
                'slide_leave',
                'slide_play')
"""Magic Events are events that are used to trigger widget actions that
are not real MPF events, rather, they're used to trigger animations from
things the slide is doing."""



__version__ = '0.54.2'
__short_version__ = '0.54'
__bcp_version__ = '1.1'
__config_version__ = '5'
__mpf_version_required__ = '0.54.0'

# pylint: disable-msg=invalid-name
version = "MPF-MC v{}".format(__version__)
'''A friendly version string for this build of MPF.'''

# pylint: disable-msg=invalid-name
extended_version = "MPF-MC v{} (config_version={}, BCP v{}, Requires MPF v{})".format(
    __version__, __config_version__, __bcp_version__, __mpf_version_required__)



from mpfmc.core.audio.sdl2 cimport *

# ---------------------------------------------------------------------------
#    Inline functions
# ---------------------------------------------------------------------------

cdef inline Uint8 lerpU8(float progress, Uint8 a, Uint8 b) nogil:
    """
    Linearly interpolate between 2 8-bit values.
    Args:
        progress: Progress (0.0 to 1.0) between the two values
        a: First 8-bit value
        b: Second 8-bit value

    Returns:
        New 8-bit value between the supplied values
    """
    return <Uint8> ((1.0 - progress) * a + progress * b)

cdef inline float in_out_quad(float progress) nogil:
    """
    A quadratic easing function used for smoother audio fading
    Args:
        progress: 0.0 to 1.0

    Notes:
        At 0.0 the output is 0.0 and at 1.0 the output is 1.0.
    """
    cdef float p
    p = progress * 2
    if p < 1:
        return 0.5 * p * p
    p -= 1.0
    return -0.5 * (p * (p - 2.0) - 1.0)




"""Interactive Media Controller"""
import asyncio
import os

from pygments.lexers import YamlLexer

from kivy.app import App
from kivy.core.window import Window
from kivy.uix.button import Button
from kivy.uix.codeinput import CodeInput
from kivy.uix.label import Label
from kivy.uix.popup import Popup
from kivy.uix.screenmanager import ScreenManager, Screen

from mpf.core.bcp.bcp import Bcp
from mpf.core.clock import ClockBase
from mpf.core.config_processor import ConfigProcessor
from mpf.core.events import EventManager
from mpf.core.mode_controller import ModeController
from mpf.core.config_validator import ConfigValidator
from mpf.file_interfaces.yaml_interface import YamlInterface

import mpfmc
from mpfmc.config_players.plugins.slide_player import MpfSlidePlayer


class InteractiveMc(App):

    """Interactive MC App."""

    # pylint: disable-msg=too-many-instance-attributes
    def __init__(self, mpf_path, machine_path, args, **kwargs):
        """Create IMC window."""
        del mpf_path
        del machine_path
        del args
        super().__init__(**kwargs)

        self.config_processor = ConfigProcessor(False, False)
        config_spec = self.config_processor.load_config_spec()
        self.config_validator = ConfigValidator(self, config_spec)
        files = [os.path.join(mpfmc.__path__[0], 'tools/interactive_mc/imcconfig.yaml')]
        self.machine_config = self.config_processor.load_config_files_with_cache(files, "machine",
                                                                                 config_spec=config_spec)
        self.machine_config['mpf'] = dict()
        self.machine_config['mpf']['allow_invalid_config_sections'] = True
        self.config = self.machine_config
        self._initialized = False
        self.options = dict(bcp=True, production=False)
        self.clock = ClockBase(self)

        # needed for bcp
        self.settings = Settings()
        self.machine_vars = {}
        self.modes = {}

        self.variables = Variables()

        self.events = EventManager(self)
        self.mode_controller = ModeController(self)
        self.bcp = Bcp(self)
        self.slide_player = MpfSlidePlayer(self)
        self.slide_player.instances['imc'] = dict()

        self.clock.loop.run_until_complete(self.events.post_queue_async("init_phase_1"))
        self.events.process_event_queue()
        self.clock.loop.run_until_complete(self.events.post_queue_async("init_phase_2"))
        self.events.process_event_queue()
        self.clock.loop.run_until_complete(self.events.post_queue_async("init_phase_3"))
        self.events.process_event_queue()
        self.clock.loop.run_until_complete(self.events.post_queue_async("init_phase_4"))
        self.events.process_event_queue()
        self.clock.loop.run_until_complete(self.events.post_queue_async("init_phase_5"))

        self.sm = ScreenManager()
        self.slide_screen = Screen(name="Slide Player")
        self.widget_screen = Screen(name="Widget Player")
        self.sm.add_widget(self.slide_screen)
        self.sm.add_widget(self.widget_screen)
        self.slide_player_code = YamlCodeInput(lexer=YamlLexer(),
                                               tab_width=4)
        self.slide_player_code.bind(on_triple_tap=self.send_slide_to_mc)

        self.slide_player_code.text = '''my_test_slide:
    widgets:
      - type: text
        text: iMC
        color: red
      - type: line
        points: 1, 1, 1, 32, 128, 32, 128, 1, 1, 1
        color: lime
      - type: rectangle
        width: 50
        height: 20
        color: yellow
'''

        self.send_button = Button(text='Send',
                                  size=(150, 60),
                                  size_hint=(None, None),
                                  background_normal='',
                                  background_color=(0, .6, 0, 1),
                                  pos=(0, 1),
                                  pos_hint={'top': 0.1, 'right': 0.95})

        self.debug_button = Button(text='Debug Dump',
                                   size=(150, 60),
                                   size_hint=(None, None),
                                   background_normal='',
                                   background_color=(0, .6, 0, 1),
                                   pos=(0, 1),
                                   pos_hint={'top': 0.1, 'left': 0.95})

        self.send_button.bind(on_press=self.send_slide_to_mc)

        self.slide_screen.add_widget(self.slide_player_code)
        self.slide_screen.add_widget(self.send_button)

        self.debug_button.bind(on_press=self.send_debug_dump_stats)
        self.slide_screen.add_widget(self.debug_button)

        self.slide_player.register_player_events(dict())

    def register_monitor(self, monitor_class, monitor):
        """Ignore register call."""
        pass

    def build(self):
        """Return screen manager."""
        return self.sm

    def send_slide_to_mc(self, value):
        """Send this slide to MC via BCP."""
        del value

        try:
            settings = YamlInterface.process(self.slide_player_code.text)
        except Exception as e:      # pylint: disable-msg=broad-except
            msg = str(e).replace('"', '\n')
            Popup(title='Error in your config',
                  content=Label(text=msg, size=(750, 350)),
                  size_hint=(None, None),
                  size=(Window.width, 400)).open()
            return

        try:
            settings = (self.slide_player.validate_config_entry(settings,
                                                                'slides'))
        except Exception as e:      # pylint: disable-msg=broad-except
            msg = str(e).replace('"', '\n')
            Popup(title='Error in your config',
                  content=Label(text=msg, size=(750, 350)),
                  size_hint=(None, None),
                  size=(Window.width, 400)).open()
            return

        if self._initialized:
            self.slide_player.clear_context('imc')
        else:
            self._initialized = True
        self.slide_player.play(settings, 'imc', 100)
        self.clock.loop.run_until_complete(asyncio.sleep(.1, loop=self.clock.loop))

    def send_debug_dump_stats(self, value):
        """Send event to MC via BCP."""
        del value
        self.bcp.interface.bcp_trigger("debug_dump_stats")
        self.clock.loop.run_until_complete(asyncio.sleep(.1, loop=self.clock.loop))


class Variables:

    """Variables of the iMC machine."""

    def __init__(self):
        """Set empty variables."""
        self.machine_vars = {}

    def set_machine_var(self, name, value):
        """Ignore changes."""
        pass


class YamlCodeInput(CodeInput):

    """Input Widget for YAML."""

    def insert_text(self, substring, from_undo=False):
        """Add text."""
        s = substring.replace('\t', '    ')
        return super().insert_text(s, from_undo=from_undo)


class Settings:

    """Empty settings."""

    @staticmethod
    def get_settings():
        """Return empty settings."""
        return {}



from kivy.uix.effectwidget import EffectBase
from kivy.properties import NumericProperty


class ReduceEffect(EffectBase):
    """GLSL effect to reduce a texture to fewer bits per color channel."""

    shades = NumericProperty(16)
    '''
    Sets the number of shades per channel to reduce it to.

    shades is a :class:`~kivy.properties.NumericProperty` and
    defaults to 16.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.do_glsl()

    def on_shades(self, *args):
        self.do_glsl()

    def do_glsl(self):
        self.glsl = reduce_glsl.format(abs(float(self.shades - 1)))


reduce_glsl = '''
        vec4 effect(vec4 color, sampler2D texture, vec2 tex_coords, vec2 coords)
        {{
        float bitDepth = {};
        vec4 outColor = vec4(floor(color.x * bitDepth) / bitDepth,
                             floor(color.y * bitDepth) / bitDepth,
                             floor(color.z * bitDepth) / bitDepth,
                             1.0);
        return outColor;
        }}
        '''


effect_cls = ReduceEffect
name = 'reduce'



from enum import Enum, unique
from typing import Optional

from mpf.core.randomizer import Randomizer
from mpfmc.core.audio.audio_exception import AudioException
from mpfmc.core.config_collection import ConfigCollection


class PlaylistCollection(ConfigCollection):

    config_section = 'playlists'
    collection = 'playlists'
    class_label = 'Playlists'

    def __init__(self, mc, collection, config_section):
        super().__init__(mc, collection, config_section)

        self._validate_handler = None

    def create_entries(self, config: dict, **kwargs) -> None:
        # Do not call base class implementation
        del kwargs

        # Loop over items in the config (localized to this collection's section)
        for name, settings in config.items():
            try:
                self[name] = self.process_config(settings)
            except (AudioException, ValueError) as ex:
                raise ValueError("An error occurred while processing the '{}' entry in "
                                 "the playlists config collection: {}".format(name, ex))

        # Validation of referenced sound assets must be completed after all
        # assets have been loaded (can use the init_done event for that)
        self._validate_handler = self.mc.events.add_handler("init_done", self._validate_sound_assets)

    def process_config(self, config: dict) -> dict:
        # processes the 'playlists' section of a config file to populate the
        # mc.playlists dict.

        # config is localized to 'playlists' section
        return self.process_playlist(config)

    def process_playlist(self, config: dict) -> dict:
        # config is localized to a single playlist settings within a list

        self.mc.config_validator.validate_config('playlists', config)

        # Clamp volume between 0 and 1
        if 'volume' in config and config['volume']:
            if config['volume'] < 0:
                config['volume'] = 0
            elif config['volume'] > 1:
                config['volume'] = 1

        return config

    def _validate_sound_assets(self, **kwargs) -> None:
        """Validate the referenced sound assets in the playlist.

        Notes:
            This must be performed after all the sound assets have been loaded.
        """
        del kwargs
        if self._validate_handler:
            self.mc.events.remove_handler(self._validate_handler)

        # bail out if there is no sound system
        if not hasattr(self.mc, "sounds"):
            return

        for name, config in self.items():
            # Validate sound settings in sounds (make sure only valid sound assets are referenced and
            # at least one sound is referenced)
            if not config["sounds"]:
                raise ValueError("The '{}' playlist does not contain any sound assets in "
                                 "its sound: section".format(name))

            for sound in config["sounds"]:
                if sound not in self.mc.sounds:
                    raise ValueError("The '{}' playlist references an invalid sound asset "
                                     "name '{}' in one of its layers".format(name, sound))


# ---------------------------------------------------------------------------
#    PlaylistInstanceStatus class
# ---------------------------------------------------------------------------
@unique
class PlaylistInstanceStatus(Enum):
    """Enumerated class containing status values for PlayfieldInstance class."""
    pending = 0
    playing = 2
    stopping = 3
    finished = 4


# ---------------------------------------------------------------------------
#    PlaylistInstance class
# ---------------------------------------------------------------------------
# pylint: disable=too-many-public-methods
class PlaylistInstance:
    """
    PlaylistInstance class represents an instance of a playlist asset.
    """

    # pylint: disable-msg=too-many-arguments
    def __init__(self, name: str, playlist: dict, track_crossfade_time: float,
                 context: Optional[str] = None, settings: Optional[dict] = None):
        """Constructor"""
        self._name = name
        if playlist is None:
            raise ValueError("Cannot create playlist instance: playlist parameter is None")

        self._context = context

        if settings is None:
            settings = {}

        # Set crossfade time that will be used for playing this instance (will either
        # use the track time or the playlist time depending upon the crossfade_mode
        # setting
        settings.setdefault('crossfade_mode', playlist['crossfade_mode'])
        if settings['crossfade_mode'] == 'use_track_setting':
            settings['crossfade_time'] = track_crossfade_time
        elif settings['crossfade_mode'] == 'use_playlist_setting':
            if playlist['crossfade_mode'] == 'use_track_setting':
                settings['crossfade_time'] = track_crossfade_time
            else:
                settings.setdefault('crossfade_time', playlist['crossfade_time'])
        else:
            settings.setdefault('crossfade_time', playlist['crossfade_time'])

        settings.setdefault('shuffle', playlist['shuffle'])
        settings.setdefault('repeat', playlist['repeat'])
        settings['sounds'] = playlist['sounds']
        settings.setdefault('events_when_played', playlist['events_when_played'])
        settings.setdefault('events_when_stopped', playlist['events_when_stopped'])
        settings.setdefault('events_when_looping', playlist['events_when_looping'])
        settings.setdefault('events_when_sound_changed', playlist['events_when_sound_changed'])
        settings.setdefault('events_when_sound_stopped', playlist['events_when_sound_stopped'])
        self._settings = settings

        self._sounds = Randomizer(self._settings['sounds'])
        self._sounds.disable_random = not self._settings['shuffle']
        self._sounds.force_all = True
        self._sounds.force_different = True
        self._sounds.loop = self._settings['repeat']

        self._current_sound_instance = None
        self._fading_sound_instance = None

    def __repr__(self):
        """String that's returned if someone prints this object"""
        return '<PlaylistInstance: {}>'.format(self.name)

    @property
    def name(self):
        return self._name

    @property
    def crossfade_time(self):
        return self._settings['crossfade_time']

    @property
    def shuffle(self):
        return self._settings['shuffle']

    @property
    def repeat(self):
        return self._settings['repeat']

    @property
    def events_when_played(self):
        return self._settings['events_when_played']

    @property
    def events_when_stopped(self):
        return self._settings['events_when_stopped']

    @property
    def events_when_looping(self):
        return self._settings['events_when_looping']

    @property
    def events_when_sound_changed(self):
        return self._settings['events_when_sound_changed']

    @property
    def events_when_sound_stopped(self):
        return self._settings['events_when_sound_stopped']

    @property
    def sounds(self):
        return self._settings['sounds']

    @property
    def end_of_playlist(self):
        if self._sounds.disable_random:
            return self._sounds.data['current_item_index'] == len(self._sounds.items)
        else:
            return len(self._sounds.data['items_sent']) == len(self._sounds.items)

    @property
    def current_sound_instance(self):
        """Return the current sound instance"""
        return self._current_sound_instance

    @current_sound_instance.setter
    def current_sound_instance(self, value):
        """Set the current sound instance"""
        self._current_sound_instance = value

    @property
    def fading_sound_instance(self):
        """Return the fading sound instance"""
        return self._fading_sound_instance

    @fading_sound_instance.setter
    def fading_sound_instance(self, value):
        """Set the fading sound instance"""
        self._fading_sound_instance = value

    @property
    def context(self):
        """The context under which this playlist was created/played."""
        return self._context

    def get_next_sound_name(self):
        """Return the name of the next sound in the playlist (advance iterator)"""
        try:
            return self._sounds.get_next()
        except StopIteration:
            return None

    def get_current_sound_name(self):
        """Return the name of the current sound in the playlist"""
        try:
            return self._sounds.get_current()
        except StopIteration:
            return None


collection_cls = PlaylistCollection



import abc


class BcpConfigPlayer(metaclass=abc.ABCMeta):

    """Base class for bcp players which play based on BCP."""

    config_file_section = None          # type: str
    show_section = None                 # type: str
    machine_collection_name = None      # type: str

    def __init__(self, machine):
        self.machine = machine
        self.instances = dict()
        self.instances['_global'] = dict()

        self.machine.events.add_handler(
            event='{}_play'.format(self.show_section),
            handler=self.play_from_trigger)

        self.machine.events.add_handler(
            event='{}_clear'.format(self.show_section),
            handler=self.clear_from_trigger)

    def _get_full_context(self, context):
        return context + "." + self.config_file_section

    def _get_instance_dict(self, context):
        return self.instances[context]

    def _reset_instance_dict(self, context):
        self.instances[context] = dict()

    def __repr__(self):
        return 'BcpConfigPlayer.{}'.format(self.show_section)

    def play_from_trigger(self, settings, context, priority, **kwargs):
        """Call play from BCP trigger."""
        if context not in self.instances:
            self.instances[context] = dict()

        self.play_element(settings=settings, context=context, priority=priority, **kwargs)

    def clear_from_trigger(self, context, **kwargs):
        """Call clear_context from BCP trigger."""
        del kwargs
        self.clear_context(context=context)

    # pylint: disable-msg=too-many-arguments
    @abc.abstractmethod
    def play_element(self, settings, element, context, calling_context, priority=0, **kwargs):
        """Directly play player."""
        # **kwargs since this is an event callback
        raise NotImplementedError

    @abc.abstractmethod
    def clear_context(self, context):
        """Clear the context."""
        raise NotImplementedError



"""Base class for remote config players."""
import abc

from mpf.config_players.device_config_player import DeviceConfigPlayer


# pylint: disable-msg=abstract-method
class McConfigPlayer(DeviceConfigPlayer, metaclass=abc.ABCMeta):

    """Remote config player which is triggered via BCP."""

    config_file_section = None
    show_section = None
    machine_collection_name = None

    def __repr__(self):
        return 'McConfigPlayer.{}'.format(self.show_section)

    def _initialise_system_wide(self, **kwargs):
        del kwargs
        # this does not call super() since the base class uses self.config
        # and the mc uses self.machine_config
        if self.machine_collection_name:
            self.device_collection = getattr(self.machine,
                                             self.machine_collection_name)
        else:
            self.device_collection = None

        # Look through the machine config for config_player sections and
        # for shows to validate and process
        if (self.config_file_section in self.machine.machine_config and
                self.machine.machine_config[self.config_file_section]):
            # Validate
            self.machine.machine_config[self.config_file_section] = (
                self.validate_config(
                    self.machine.machine_config[self.config_file_section]))

            self.register_player_events(
                self.machine.machine_config[self.config_file_section])

        self.machine.events.add_handler(
            event='{}_play'.format(self.show_section),
            handler=self.play_from_trigger)

        self.machine.events.add_handler(
            event='{}_clear'.format(self.show_section),
            handler=self.clear_from_trigger)

    def play_from_trigger(self, settings, context, priority, calling_context, **kwargs):
        """Call play from BCP trigger."""
        if context not in self.instances:
            self.instances[context] = dict()
        if self.config_file_section not in self.instances[context]:
            self.instances[context][self.config_file_section] = dict()

        self.play(settings=settings, context=context, calling_context=calling_context, priority=priority, **kwargs)

    def clear_from_trigger(self, context, **kwargs):
        """Call clear_context from BCP trigger."""
        del kwargs
        self.clear_context(context=context)



from kivy.properties import OptionProperty

from mpfmc.uix.transitions import MpfTransition


class MoveOutTransition(MpfTransition):
    """Move Out Transition. The current slide moves out of the frame to
    reveal the new slide underneath it"""

    direction = OptionProperty('left', options=('left', 'right', 'top',
                                                'bottom'))

    def start(self, manager):

        super().start(manager)

        # Need to move the new slide so it's below the old one.
        manager.screens.insert(1, manager.screens.pop(0))

    def on_progress(self, progression):

        s_in, s_out, width, height, progress = self.get_vars(progression)

        del s_in

        direction = self.direction

        if direction == 'left':
            s_out.x = 0 - width * progress

        elif direction == 'right':
            s_out.x = width * progress

        elif direction == 'top':
            s_out.y = 0 - height * progress

        elif direction == 'bottom':
            s_out.y = height * progress


transition_cls = MoveOutTransition
name = 'move_out'



from kivy.core.video import Video
from kivy.core.video.video_null import VideoNull
from kivy.properties import AliasProperty

from mpf.core.assets import AssetPool
from mpfmc.assets.mc_asset import McAsset


class VideoPool(AssetPool):

    def __repr__(self):
        return '<VideoPool: {}>'.format(self.name)

    @property
    def video(self):
        return self.asset


class VideoWrapper(Video):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.register_event_type('on_play')
        self.register_event_type('on_stop')

    def on_play(self):
        pass

    def on_stop(self):
        pass

    def stop(self):
        super().stop()
        self.dispatch('on_stop')

    def play(self):
        super().play()
        self.dispatch('on_play')


class VideoAsset(McAsset):

    attribute = 'videos'
    path_string = 'videos'
    config_section = 'videos'
    extensions = ('mkv', 'avi', 'mpg', 'mp4', 'm4v', 'mov')
    class_priority = 100
    pool_config_section = 'video_pools'
    asset_group_class = VideoPool

    def __init__(self, mc, name, file, config):
        self._video = None      # type: VideoWrapper
        super().__init__(mc, name, file, config)

        # Setup events to post when video state changes
        self._events_when_played = list()
        self._events_when_stopped = list()

        if self.config['events_when_played']:
            self._events_when_played = self.config['events_when_played']

        if self.config['events_when_stopped']:
            self._events_when_stopped = self.config['events_when_stopped']

    @property
    def video(self):
        return self._video

    def do_load(self):
        # For videos, we need them to load in the main thread, so we do not
        # load them here and load them via is_loaded() below.
        pass

    def _do_unload(self):
        if self._video:
            self._video.stop()
            self._video.unload()
            self._video = None

    def set_end_behavior(self, eos='stop'):
        assert eos in ('loop', 'pause', 'stop')
        self._video.eos = eos

    def is_loaded(self):
        self.loading = False
        self.loaded = True
        self.unloading = False
        self._video = VideoWrapper(filename=self.file)
        self._video.bind(on_load=self._check_duration,
                         on_play=self.on_play,
                         on_stop=self.on_stop)

        if isinstance(self._video, VideoNull):
            raise AssertionError("Kivy cannot load video {} because there is no provider.".format(self.file))

        self._call_callbacks()

    def _check_duration(self, instance):
        del instance
        if self._video and self._video.duration <= 0:
            raise ValueError(
                "Video file {} was loaded, but seems to have no content. Check"
                " to make sure you have the proper Gstreamer plugins for the "
                "codec this video needs".format(self.file))

    def on_play(self, *args):
        del args

        if self._events_when_played:
            for event in self._events_when_played:
                self.machine.post_mc_native_event(event)

    def on_stop(self, *args):
        del args

        if self._events_when_stopped:
            for event in self._events_when_stopped:
                self.machine.post_mc_native_event(event)

    #
    # Properties
    #

    def _get_position(self):
        try:
            return self._video.position
        except AttributeError:
            return 0

    def _set_position(self, pos):
        # position in secs
        try:
            self._video.position = pos  # noqa
        except AttributeError:
            pass

    position = AliasProperty(_get_position, _set_position)
    '''Position of the video between 0 and :attr:`duration`. The position
    defaults to -1 and is set to a real position when the video is loaded.

    :attr:`position` is a :class:`~kivy.properties.NumericProperty` and
    defaults to -1.
    '''

    def _get_duration(self):
        try:
            return self._video.duration
        except AttributeError:
            return 0

    duration = AliasProperty(_get_duration, None)
    '''Duration of the video. The duration defaults to -1, and is set to a real
    duration when the video is loaded.

    :attr:`duration` is a :class:`~kivy.properties.NumericProperty` and
    defaults to -1.
    '''

    def _get_volume(self):
        try:
            return self._video.volume
        except AttributeError:
            return 0

    def _set_volume(self, volume):
        # float 0.0 - 1.0
        try:
            self._video.volume = volume     # noqa
        except AttributeError:
            pass

    volume = AliasProperty(_get_volume, _set_volume)
    '''Volume of the video, in the range 0-1. 1 means full volume, 0
    means mute.

    :attr:`volume` is a :class:`~kivy.properties.NumericProperty` and defaults
    to 1.
    '''

    def _get_state(self):
        return self._video.state

    state = AliasProperty(_get_state, None)
    '''String, indicates whether to play, pause, or stop the video::
    '''



import logging
import os
import sys

import mpf.core

os.environ['KIVY_NO_FILELOG'] = '1'
os.environ['KIVY_NO_CONSOLELOG'] = '1'
os.environ["KIVY_NO_ARGS"] = "1"

from queue import Queue

import time
from kivy.config import Config
from kivy.logger import Logger
from kivy.base import runTouchApp, EventLoop
from kivy.clock import Clock
from kivy.uix.widget import Widget as KivyWidget

for handler in Logger.handlers:
    Logger.removeHandler(handler)

sys.stderr = sys.__stderr__

import mpfmc
import mpfmc.core
from mpf.tests.MpfBcpTestCase import MockBcpClient
from mpf.tests.MpfTestCase import MpfTestCase, patch, UnitTestConfigLoader


class TestBcpClient(MockBcpClient):
    def __init__(self, machine, name, bcp):
        super().__init__(machine, name, bcp)
        self.queue = Queue()
        self.exit_on_close = False

        self.fps = 30

        self._start_time = time.time()
        Clock._start_tick = self._start_time
        Clock._last_tick = self._start_time
        Clock.time = self._mc_time
        Clock._events = [[] for i in range(256)]
        with patch("mpfmc.core.bcp_processor.BCPServer"):
            self._start_mc()
        self.mc_task = self.machine.clock.schedule_interval(self._run_mc, 1 / self.fps)

        bcp_mc = self.mc.bcp_processor
        bcp_mc.send = self.receive
        self.queue = bcp_mc.receive_queue
        self.mc.bcp_processor.enabled = True
        self.mc.bcp_client_connected = True
        self.mc.events.post("client_connected")

    def get_absolute_machine_path(self):
        # creates an absolute path based on machine_path
        return self.machine.machine_path

    def get_options(self):
        return dict(machine_path=self.get_absolute_machine_path(),
                    mcconfigfile='mcconfig.yaml',
                    production=False,
                    configfile=self.machine.options['configfile'],
                    no_load_cache=False,
                    create_config_cache=True,
                    bcp=False)

    def preprocess_config(self, config):
        # TODO this method is copied from the mc.py launcher. Prob a better way
        kivy_config = config['kivy_config']

        try:
            kivy_config['graphics'].update(config['displays']['window'])
        except KeyError:
            pass

        try:
            kivy_config['graphics'].update(config['window'])
        except KeyError:
            pass

        if 'top' in kivy_config['graphics'] and 'left' in kivy_config['graphics']:
            kivy_config['graphics']['position'] = 'custom'

        for section, settings in kivy_config.items():
            for k, v in settings.items():
                try:
                    if k in Config[section]:
                        Config.set(section, k, v)
                except KeyError:
                    continue

    def _start_app_as_slave(self):
        # from app::run
        if not self.mc.built:
            self.mc.load_config()
            self.mc.load_kv(filename=self.mc.kv_file)
            root = self.mc.build()
            if root:
                self.mc.root = root
        if self.mc.root:
            if not isinstance(self.mc.root, KivyWidget):
                Logger.critical('App.root must be an _instance_ of Kivy Widget')
                raise Exception('Invalid instance in App.root')
            from kivy.core.window import Window
            Window.add_widget(self.mc.root)

        # Check if the window is already created
        from kivy.base import EventLoop
        window = EventLoop.window
        if window:
            self.mc._app_window = window
            window.set_title(self.mc.get_application_name())
            icon = self.mc.get_application_icon()
            if icon:
                window.set_icon(icon)
            self.mc._install_settings_keys(window)
        else:
            Logger.critical("Application: No window is created."
                            " Terminating application run.")
            return

        self.mc.dispatch('on_start')
        runTouchApp(slave=True)  # change is here

        while not self.mc.is_init_done.is_set():
            EventLoop.idle()

    def _start_mc(self):
        from mpfmc.core.mc import MpfMc

        # prevent sleep in clock
        Clock._max_fps = 0

        machine_path = self.get_absolute_machine_path()

        config_loader = UnitTestConfigLoader(machine_path, self.machine.options['configfile'], {}, {}, {})

        config = config_loader.load_mc_config()

        self.mc = MpfMc(config=config, options=self.get_options())

        from kivy.core.window import Window
        Window.create_window()
        Window.canvas.clear()

        self._start_app_as_slave()

    def _mc_time(self):
        return self._start_time + self.machine.clock.loop._time

    def _run_mc(self):
        EventLoop.idle()

    def stop(self):
        self.mc.stop()
        self.machine.clock.unschedule(self.mc_task)

    def send(self, bcp_command, kwargs):
        self.queue.put((bcp_command, kwargs))

    def receive(self, bcp_command, callback=None, rawbytes=None, **kwargs):
        if rawbytes:
            kwargs['rawbytes'] = rawbytes
        self.receive_queue.put_nowait((bcp_command, kwargs))

        if callback:
            callback()


class MpfIntegrationTestCase(MpfTestCase):

    fps = 30

    def get_use_bcp(self):
        return True

    def get_absolute_machine_path(self):
        # creates an absolute path based on machine_path
        return os.path.abspath(os.path.join(
            mpfmc.core.__path__[0], os.pardir, self.get_machine_path()))

    def get_enable_plugins(self):
        return True

    def mock_mc_event(self, event_name):
        """Configure an event to be mocked.

        Same as mock_event but for mc in integration test.
        """
        self._mc_events[event_name] = 0
        self.mc.events.remove_handler_by_event(event=event_name, handler=self._mock_mc_event_handler)
        self.mc.events.add_handler(event=event_name,
                                   handler=self._mock_mc_event_handler,
                                   event_name=event_name)

    def _mock_mc_event_handler(self, event_name, **kwargs):
        self._last_mc_event_kwargs[event_name] = kwargs
        self._mc_events[event_name] += 1

    def assertMcEventNotCalled(self, event_name):
        """Assert that event was not called.

        Same as mock_event but for mc in integration test.
        """
        if event_name not in self._mc_events:
            raise AssertionError("Event {} not mocked.".format(event_name))

        if self._mc_events[event_name] != 0:
            raise AssertionError("Event {} was called {} times.".format(event_name, self._mc_events[event_name]))

    def assertMcEventCalled(self, event_name, times=None):
        """Assert that event was called.

        Same as mock_event but for mc in integration test.
        """
        if event_name not in self._mc_events:
            raise AssertionError("Event {} not mocked.".format(event_name))

        if self._mc_events[event_name] == 0 and times != 0:
            raise AssertionError("Event {} was not called.".format(event_name))

        if times is not None and self._mc_events[event_name] != times:
            raise AssertionError("Event {} was called {} instead of {}.".format(
                event_name, self._mc_events[event_name], times))

    def __init__(self, methodName):
        super().__init__(methodName)
        self._mc_events = {}
        self._last_mc_event_kwargs = {}
        self.console_logger = None
        try:
            del self.machine_config_patches['mpf']['plugins']
        except KeyError:
            pass
        self.machine_config_patches['bcp'] = \
            {"connections": {"local_display": {"type": "mpfmc.tests.MpfIntegrationTestCase.TestBcpClient"}}}
        self.machine_config_patches['bcp']['servers'] = []
        self.expected_duration = 60

    def setUp(self):
        if self.unittest_verbosity() > 1:
            self.console_logger = logging.StreamHandler()
            self.console_logger.setLevel(logging.DEBUG)

            # set a format which is simpler for console use
            formatter = logging.Formatter('%(name)s: %(message)s')

            # tell the handler to use this format
            self.console_logger.setFormatter(formatter)

            # add the handler to the root logger
            logging.getLogger('').addHandler(self.console_logger)
        super().setUp()

        client = self.machine.bcp.transport.get_named_client("local_display")
        self.mc = client.mc
        self.advance_time_and_run()

    def tearDown(self):
        super().tearDown()
        EventLoop.close()

        if self.console_logger:
            logging.getLogger('').removeFilter(self.console_logger)
            self.console_logger = None



from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase
from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase


class TestService(MpfIntegrationTestCase, MpfSlideTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'integration/machine_files/service_mode/'

    def test_service_slides(self):
        # open door
        self.hit_switch_and_run("s_door_open", 1)
        self.assertModeRunning("attract")
        self.assertSlideOnTop("service_door_open")
        self.assertTextOnTopSlide("Coil Power Off")

        # enter
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")
        self.assertTextOnTopSlide("Diagnostics Menu")

        # enter diagnostics menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Switch Menu")

        # enter switch menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Switch Edge Test")

        # enter switch test
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_switch_test")

        self.hit_and_release_switch("s_test")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("s_test")
        self.assertTextOnTopSlide("The test switch label")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")
        self.assertTextOnTopSlide("Switch Edge Test")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")
        self.assertTextOnTopSlide("Switch Menu")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Coil Menu")

        # enter coil menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Single Coil Test")

        # enter coil test
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_coil_test")
        self.assertTextOnTopSlide("c_test")
        self.assertTextOnTopSlide("First coil")
        self.assertTextOnTopSlide("Coil Power Off")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")
        self.assertTextOnTopSlide("Single Coil Test")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")
        self.assertTextOnTopSlide("Coil Menu")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Light Menu")

        # enter light menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Single Light Test")

        # enter light test
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_light_test")
        self.assertTextOnTopSlide("l_light1")
        self.assertTextOnTopSlide("First light")
        self.assertTextOnTopSlide("1/white")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")
        self.assertTextOnTopSlide("Single Light Test")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")
        self.assertTextOnTopSlide("Light Menu")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")
        self.assertTextOnTopSlide("Diagnostics Menu")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Audits Menu")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Adjustments Menu")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Utilities Menu")

        # enter util menu
        self.hit_and_release_switch("s_service_enter")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Reset Menu")

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run(.1)
        self.assertTextOnTopSlide("Software Update")

        # close door
        self.release_switch_and_run("s_door_open", 1)
        self.assertTextNotOnTopSlide("Coil Power Off")

        # exit
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)
        self.assertSlideOnTop("service_menu")

        # exit service menu
        self.hit_and_release_switch("s_service_esc")
        self.advance_time_and_run(.1)

        if not self.mc.sound_system or not self.mc.sound_system.audio_interface:
            return

        # test volume
        self.assertEqual(0.8, self.machine.variables.get_machine_var("master_volume"))

        self.hit_and_release_switch("s_service_up")
        self.advance_time_and_run(.1)
        self.assertEqual(0.84, self.machine.variables.get_machine_var("master_volume"))



from mpf.config_players.plugin_player import PluginPlayer


class MpfSoundLoopPlayer(PluginPlayer):
    """Base class for part of the sound loop player which runs as part of MPF.

    Note: This class is loaded by MPF and everything in it is in the context of
    MPF, not the mpf-mc. MPF finds this instance because the mpf-mc setup.py
    has the following entry_point configured:

        sound_loop_player=mpfmc.config_players.sound_loop_player:register_with_mpf

    """
    config_file_section = 'sound_loop_player'
    show_section = 'sound_loop_sets'

    def _validate_config_item(self, device, device_settings):
        """Validates the config when in a show or in a player"""

        # event contains the event name that triggers the sound_loop_player action
        # Validate the settings against the config spec

        # First validate the action item (since it will be used to validate the rest
        # of the config)
        if 'action' in device_settings:
            device_settings['action'] = self.machine.config_validator.validate_config_item(
                self.machine.config_validator.config_spec['sound_loop_player']['action'],
                'sound_loop_player:{}'.format(device),
                device_settings['action'])
        else:
            device_settings['action'] = self.machine.config_validator.validate_config_item(
                self.machine.config_validator.config_spec['sound_loop_player']['action'],
                'sound_loop_player:{}'.format(device))

        validated_settings = self.machine.config_validator.validate_config(
            'sound_loop_player_actions:{}'.format(device_settings['action']).lower(),
            device_settings)

        # Remove any items from the settings that were not explicitly provided in the
        # sound_loop_player config section (only want to override sound settings explicitly
        # and not with any default values).  The default values for these items are not
        # legal values and therefore we know the user did not provide them.
        if 'volume' in validated_settings and validated_settings['volume'] is None:
            del validated_settings['volume']
        if 'fade_in' in validated_settings and validated_settings['fade_in'] is None:
            del validated_settings['fade_in']
        if 'fade_out' in validated_settings and validated_settings['fade_out'] is None:
            del validated_settings['fade_out']
        if 'events_when_played' in validated_settings and len(validated_settings['events_when_played']) == 1 and \
                validated_settings['events_when_played'][0] == 'use_sound_loop_setting':
            del validated_settings['events_when_played']
        if 'events_when_stopped' in validated_settings and len(validated_settings['events_when_stopped']) == 1 and \
                validated_settings['events_when_stopped'][0] == 'use_sound_loop_setting':
            del validated_settings['events_when_stopped']
        if 'events_when_looping' in validated_settings and len(validated_settings['events_when_looping']) == 1 and \
                validated_settings['events_when_looping'][0] == 'use_sound_loop_setting':
            del validated_settings['events_when_looping']
        if 'mode_end_action' in validated_settings and (validated_settings['mode_end_action'] is None or
                                                        validated_settings['mode_end_action'] ==
                                                        'use_sound_loop_setting'):
            del validated_settings['mode_end_action']

        validated_dict = dict()
        validated_dict[device] = validated_settings
        return validated_dict


player_cls = MpfSoundLoopPlayer


def register_with_mpf(machine):
    """Registers the sound loop player plug-in with MPF"""
    return 'sound_loop', MpfSoundLoopPlayer(machine)



#!python
#cython: embedsignature=True, language_level=3

from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
cimport cpython.pycapsule as pycapsule
import logging
import os

from mpfmc.core.audio.audio_exception import AudioException
from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *
from mpfmc.core.audio.sound_file cimport *


# ---------------------------------------------------------------------------
#    SoundFile class
# ---------------------------------------------------------------------------
cdef class SoundFile:
    """SoundFile is the base class for wrapper classes used to manage sound sample data."""

    def __init__(self, str file_name, object audio_callback_data):
        self.log = logging.getLogger("SoundFile")
        self.file_name = file_name
        self.callback_data = <AudioCallbackData*>pycapsule.PyCapsule_GetPointer(audio_callback_data, NULL)
        self.sample.duration = 0

    def __repr__(self):
        return '<SoundFile>'

    def load(self):
        """Load the sound file"""
        raise NotImplementedError("Must be implemented in derived class")

    def unload(self):
        """Unload the sound file"""
        raise NotImplementedError("Must be implemented in derived class")

    @property
    def duration(self):
        """Return the duration of the sound file"""
        return self.sample.duration


# ---------------------------------------------------------------------------
#    SoundMemoryFile class
# ---------------------------------------------------------------------------
cdef class SoundMemoryFile(SoundFile):
    """SoundMemoryFile is a wrapper class to manage sound sample data stored
    in memory."""

    def __init__(self, str file_name, object audio_callback_data):
        # IMPORTANT: Call super class init function
        super().__init__(file_name, audio_callback_data)
        self.log = logging.getLogger("SoundMemoryFile")
        self.sample.type = sound_type_memory
        self.sample.data.memory = <SampleMemory*>PyMem_Malloc(sizeof(SampleMemory))
        self.sample.data.memory.data = NULL
        self.sample.data.memory.size = 0

        self.load()

    def __dealloc__(self):
        self.unload()
        if self.sample.data.memory != NULL:
            PyMem_Free(self.sample.data.memory)
            self.sample.data.memory = NULL

    def __repr__(self):
        if self.loaded:
            return '<SoundMemoryFile({}, Loaded=True, sample_duration={}s)>'.format(self.file_name, self.sample.duration)
        else:
            return "<SoundMemoryFile({}, Loaded=False)>".format(self.file_name)

    def load(self):
        """Loads the sound into memory using the most appropriate library for the format."""
        cdef Mix_Chunk *chunk

        if self.loaded:
            return

        if not os.path.isfile(self.file_name):
            raise AudioException('Could not locate file ' + self.file_name)

        # Load the audio file (will be converted to current sample output format)
        chunk = Mix_LoadWAV(self.file_name.encode('utf-8'))
        if chunk == NULL:
            msg = "Could not load sound file {} due to an error: {}".format(self.file_name, SDL_GetError())
            raise AudioException(msg)

        # Save the loaded sample data
        self.sample.data.memory.size = <gsize>chunk.alen
        self.sample.data.memory.data = <gpointer>chunk.abuf

        # Set the sample size (bytes) and  duration (seconds)
        self.sample.size = self.sample.data.memory.size
        self.sample.duration = self.sample.data.memory.size / self.callback_data.seconds_to_bytes_factor

        # Free the chunk (sample memory will not be freed).  The chunk was allocated using SDL_malloc in the
        # SDL_Mixer library.  We do not want to use Mix_Free or the sample data will be freed.  Instead, we
        # can just free the Mix_Chunk structure using SDL_free and the sample buffer will remain intact. The
        # sample memory must be freed later when this object is deallocated.
        SDL_free(chunk)

        self.log.debug('Loaded file: %s Sample duration: %s',
                       self.file_name, self.sample.duration)

    def unload(self):
        """Unloads the sample data from memory"""
        if self.sample.data.memory.data != NULL:
            SDL_free(<void*>self.sample.data.memory.data)

        self.sample.data.memory.data = NULL
        self.sample.data.memory.size = 0

    @property
    def loaded(self):
        """Returns whether or not the sound file data is loaded in memory"""
        return self.sample.data.memory.data != NULL and self.sample.data.memory.size > 0


# ---------------------------------------------------------------------------
#    SoundStreamingFile class
# ---------------------------------------------------------------------------
cdef class SoundStreamingFile(SoundFile):
    """SoundStreamingFile is a wrapper class to manage streaming sound sample data."""

    def __cinit__(self, *args, **kwargs):
        """C constructor"""
        self.pipeline = NULL
        self.bus = NULL
        self.bus_message_handler_id = 0

    def __init__(self, str file_name, object audio_callback_data):
        # IMPORTANT: Call super class init function
        super().__init__(file_name, audio_callback_data)
        self.log = logging.getLogger("SoundStreamingFile")

        self.sample.type = sound_type_streaming
        self.sample.data.stream = <SampleStream*>PyMem_Malloc(sizeof(SampleStream))
        self.sample.data.stream.pipeline = NULL
        self.sample.data.stream.sink = NULL
        self.sample.data.stream.sample = NULL
        self.sample.data.stream.buffer = NULL
        self.sample.data.stream.map_contains_valid_sample_data = 0
        self.sample.data.stream.map_buffer_pos = 0
        self.sample.data.stream.null_buffer_count = 0

        self.load()

    def __dealloc__(self):
        if self.sample.data.stream != NULL:
            PyMem_Free(self.sample.data.stream)
            self.sample.data.stream = NULL

    def __repr__(self):
        if self.loaded:
            return '<SoundStreamingFile({}, Loaded=True, sample_duration={}s)>'.format(self.file_name,self.sample.duration)
        return "<SoundStreamingFile({}, Loaded=False)>".format(self.file_name)

    def _gst_init(self):
        if gst_is_initialized():
            return True
        cdef int argc = 0
        cdef char **argv = NULL
        cdef GError *error
        if not gst_init_check(&argc, &argv, &error):
            msg = 'Unable to initialize gstreamer: code={} message={}'.format(
                    error.code, <bytes>error.message)
            raise AudioException(msg)

    def _destroy_pipeline(self):
        """Destroys the GStreamer pipeline"""
        """Destroys the current pipeline"""
        cdef GstState current_state, pending_state

        if self.bus != NULL and self.bus_message_handler_id != 0:
            c_signal_disconnect(<GstElement*>self.bus, self.bus_message_handler_id)
            self.bus_message_handler_id = 0

        if self.pipeline != NULL:
            # the state changes are async. if we want to guarantee that the
            # state is set to NULL, we need to query it. We also put a 5s
            # timeout for safety, but normally, nobody should hit it.
            with nogil:
                gst_element_set_state(self.pipeline, GST_STATE_NULL)
                gst_element_get_state(self.pipeline, &current_state,
                        &pending_state, <GstClockTime>5e9)
            gst_object_unref(self.pipeline)

        if self.bus != NULL:
            gst_object_unref(self.bus)

        self.bus = NULL
        self.pipeline = NULL

    def _construct_pipeline(self):
        """Creates the GStreamer pipeline used to stream the sound data"""
        cdef GError *error
        cdef GstSample *sample
        cdef gint64 size = 0

        # Pipeline structure: uridecodebin --> audioconvert --> audioresample --> appsink

        # If the pipeline has already been created, delete it
        if self.pipeline != NULL:
            self._destroy_pipeline()

        # Pipeline structure: uridecodebin --> audioconvert --> audioresample --> appsink

        # Create GStreamer pipeline with the specified caps (from a string)
        file_path = 'file:///' + self.file_name.replace('\\', '/')
        if SDL_AUDIO_ISLITTLEENDIAN(self.callback_data.format):
            audio_format = "S16LE"
        else:
            audio_format = "S16BE"
        pipeline_string = 'uridecodebin uri="{}" ! audioconvert ! audioresample ! appsink name=sink caps="audio/x-raw,rate={},channels={},format={},layout=interleaved" sync=true blocksize={}'.format(
            file_path, str(self.callback_data.sample_rate), str(self.callback_data.channels), audio_format, self.callback_data.buffer_size)

        error = NULL
        self.pipeline = gst_parse_launch(pipeline_string.encode('utf-8'), &error)

        if error != NULL:
            msg = 'Unable to create a GStreamer pipeline: code={} message={}'.format(error.code, <bytes>error.message)
            raise AudioException(msg)

        # Get the pipeline bus (the bus allows applications to receive pipeline messages)
        self.bus = gst_pipeline_get_bus(<GstPipeline*>self.pipeline)
        if self.bus == NULL:
            raise AudioException('Unable to get bus from the pipeline')

        # Enable pipeline messages and callback message handler
        #gst_bus_enable_sync_message_emission(self.bus)
        #self.bus_message_handler_id = c_bus_connect_message(self.bus, _on_gst_bus_message, <void*>self.audio_interface)

        # Get sink
        self.sink = gst_bin_get_by_name(<GstBin*>self.pipeline, "sink")

        # Set to PAUSED to make the first frame arrive in the sink
        ret = gst_element_set_state(self.pipeline, GST_STATE_PAUSED)

        # Get the preroll sample (forces the code to wait until the sample has been completely loaded
        # which is necessary to retrieve the duration).
        sample = c_appsink_pull_preroll(self.sink)
        if sample != NULL:
            gst_sample_unref(sample)

        # Get size of audio file (in bytes)
        if not gst_element_query_duration(self.sink, GST_FORMAT_BYTES , &size):
            size = 0

        # Store length and duration (seconds)
        self.sample.size = size
        self.sample.duration = <double>size / self.callback_data.seconds_to_bytes_factor

        # The pipeline should now be ready to play.  Store the pointers to the pipeline
        # and appsink in the SampleStream struct for use in the application.
        self.sample.data.stream.pipeline = self.pipeline
        self.sample.data.stream.sink = self.sink

    def load(self):
        """Loads the sound into memory using GStreamer"""

        #if self.loaded:
        #    return

        self._gst_init()
        self._construct_pipeline()

        self.log.debug('Loaded file: %s Sample duration: %s',
                       self.file_name, self.sample.duration)

    def unload(self):
        """Unloads the sample data from memory"""

        # Done with the streaming buffer, release references to it
        if self.sample.data.stream.map_contains_valid_sample_data:
            gst_buffer_unmap(self.sample.data.stream.buffer, &self.sample.data.stream.map_info)
            gst_sample_unref(self.sample.data.stream.sample)

            self.sample.data.stream.buffer = NULL
            self.sample.data.stream.sample = NULL
            self.sample.data.stream.map_buffer_pos = 0
            self.sample.data.stream.map_contains_valid_sample_data = 0

        # Cleanup the streaming pipeline
        gst_element_set_state(self.pipeline, GST_STATE_NULL)
        gst_object_unref(self.pipeline)

    @property
    def loaded(self):
        """Returns whether or not the sound file data is loaded in memory"""
        return self.sample.data.stream != NULL and self.sample.data.stream.pipeline != NULL and self.sample.data.stream.sink != NULL



from mpfmc.core.config_collection import ConfigCollection


class WidgetStyleCollection(ConfigCollection):

    config_section = 'widget_styles'
    collection = 'widget_styles'
    class_label = 'WidgetStyles'

    def process_config(self, config: dict):
        # config is localized to the 'widget_styles' section
        self.mc.config_validator.validate_config('widget_styles', config,
                                                 add_missing_keys=False)

        return config


collection_cls = WidgetStyleCollection



from mpf.tests.MpfMachineTestCase import BaseMpfMachineTestCase
from mpfmc.tests.MpfIntegrationTestCase import MpfIntegrationTestCase
from mpfmc.tests.MpfSlideTestCase import MpfSlideTestCase


class FullMachineTestCase(BaseMpfMachineTestCase, MpfIntegrationTestCase, MpfSlideTestCase):

    """MPF + MC machine test case."""

    fps = 3

    pass



from mpfmc.uix.display import Display
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestDisplayMultiple(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/display'

    def get_config_file(self):
        return 'test_display_multiple.yaml'

    def test_mc_display(self):
        # Make sure a multiple displays are loaded properly:

        self.assertIn('window', self.mc.displays)
        self.assertTrue(isinstance(self.mc.displays['window'], Display))
        self.assertEqual(self.mc.displays['window'].size, [401, 301])

        self.assertIn('display2', self.mc.displays)
        self.assertTrue(isinstance(self.mc.displays['display2'], Display))
        self.assertEqual(self.mc.displays['display2'].size, [402, 302])

        self.assertEqual(self.mc.targets['display2'],
                         self.mc.displays['display2'])



"""Contains classes which are used to process config files for the media
controller.

"""

import logging
from mpf.core.config_processor import ConfigProcessor as ConfigProcessorBase

# from mpfmc.uix.display import Display is imported deeper in this file
# we need it there because if we do a standard import, the Clock is created
# before we have a chance to read the config and set the maxfps


class ConfigProcessor(ConfigProcessorBase):

    """Reads the config for mc."""

    config_spec = None

    def __init__(self, machine):
        super().__init__(False, False)
        self.mc = machine
        self.machine = machine
        self.system_config = self.mc.machine_config['mpf-mc']
        self.machine_sections = None
        self.mode_sections = None

        self.machine_sections = dict()
        self.mode_sections = dict()

        # process mode-based and machine-wide configs
        self.register_load_methods()

        self.mc.events.add_handler('init_phase_1', self._init)

        # todo need to clean this up
        try:
            self.process_displays(config=self.mc.machine_config['displays'])
        except KeyError:
            pass

        if not self.mc.displays:
            from mpfmc.uix.display import Display
            Display.create_default_display(self.mc)

    def _init(self, **kwargs):
        del kwargs
        self.process_config_file(section_dict=self.machine_sections,
                                 config=self.mc.machine_config)

    def register_load_methods(self):
        """Register load method for modes."""
        for section in self.mode_sections:
            self.machine.mode_controller.register_load_method(
                load_method=self.process_mode_config,
                config_section_name=section, section=section)

    def process_config_file(self, section_dict, config):
        """Called to process a config file (can be a mode or machine config)."""
        for section in section_dict:
            if section in section_dict and section in config:
                self.process_localized_config_section(config=config[section],
                                                      section=section)

    def process_localized_config_section(self, config, section):
        """Process a single key within a config file.

        Args:
            config: The subsection of a config dict to process
            section: The name of the section, either 'scripts' or 'shows'.

        """
        self.machine_sections[section](config)

    def process_mode_config(self, config, mode, mode_path, section, **kwargs):
        """Process a mode config."""
        del mode
        del mode_path
        del kwargs
        self.process_localized_config_section(config, section)

    def process_displays(self, config):
        # config is localized to 'displays' section
        for display, settings in config.items():
            self.mc.displays[display] = self.create_display(display, settings)

    def create_display(self, name, config):
        # config is localized display settings
        from mpfmc.uix.display import Display
        return Display(self.mc, name, **self.machine.config_validator.validate_config('displays', config))



import logging

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from unittest.mock import MagicMock, ANY

try:
    from mpfmc.core.audio import SoundSystem
    from mpfmc.assets.sound import SoundStealingMethod
except ImportError:
    SoundSystem = None
    SoundStealingMethod = None
    logging.warning("mpfmc.core.audio library could not be loaded. Audio "
                    "features will not be available")


class TestAudioSoundLoop(MpfMcTestCase):
    """
    """
    def get_machine_path(self):
        return 'tests/machine_files/audio'

    def get_config_file(self):
        return 'test_audio_sound_loop.yaml'

    def test_sound_loop_track(self):
        """ Tests the Sound Loop track type and its associated assets"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudioSoundLoop')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudioSoundLoop')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Check sound loop track
        track_loops = interface.get_track_by_name("loops")
        self.assertIsNotNone(track_loops)
        self.assertEqual(track_loops.name, "loops")
        self.assertAlmostEqual(track_loops.volume, 0.6, 1)

        # /sounds/loops
        self.assertTrue(hasattr(self.mc, 'sounds'))
        self.assertIn('hihat', self.mc.sounds)
        self.assertIn('kick', self.mc.sounds)
        self.assertIn('kick2', self.mc.sounds)

        # Sound loop sets
        self.assertTrue(hasattr(self.mc, 'sound_loop_sets'))
        self.assertIn('hi_hat', self.mc.sound_loop_sets)
        self.assertIn('basic_beat', self.mc.sound_loop_sets)
        self.assertIn('basic_beat2', self.mc.sound_loop_sets)

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        # Test sound_loop_player
        self.advance_time()
        status = track_loops.get_status()
        self.assertEqual(0, len(status))

        self.mc.events.post('play_hi_hat')
        self.advance_real_time(1)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['sound_id'])
        self.assertEqual(130.0, status[0]['tempo'])

        # Ensure sound_loop_set.events_when_played is working properly (send event when a sound_loop_set is played)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='hi_hat_played')

        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='hi_hat_looping')
        self.mc.bcp_processor.send.reset_mock()

        self.mc.events.post('play_basic_beat')
        self.advance_real_time(0.1)
        status = track_loops.get_status()
        self.assertEqual(2, len(status))
        self.assertEqual('delayed', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['kick'].id, status[0]['sound_id'])
        self.assertEqual('playing', status[1]['status'])
        self.assertEqual(325660, status[1]['length'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[1]['sound_id'])
        self.assertEqual(status[0]['start_delay_samples_remaining'], status[1]['stop_loop_samples_remaining'])

        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='hi_hat_stopped')
        self.mc.bcp_processor.send.assert_any_call('trigger', name='basic_beat_played')
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['kick'].id, status[0]['sound_id'])

        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='basic_beat_looping')
        self.mc.bcp_processor.send.reset_mock()
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['kick'].id, status[0]['sound_id'])

        self.mc.events.post('play_hi_hat')
        self.advance_real_time(0.1)
        status = track_loops.get_status()
        self.assertEqual(2, len(status))
        self.assertEqual('delayed', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['sound_id'])
        self.assertEqual('playing', status[1]['status'])
        self.assertEqual(325660, status[1]['length'])
        self.assertEqual(self.mc.sounds['kick'].id, status[1]['sound_id'])
        self.assertEqual(status[0]['start_delay_samples_remaining'], status[1]['stop_loop_samples_remaining'])

        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='basic_beat_stopped')
        self.mc.bcp_processor.send.assert_any_call('trigger', name='hi_hat_played')
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['sound_id'])

        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='hi_hat_looping')
        self.mc.bcp_processor.send.reset_mock()
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['sound_id'])

        self.mc.events.post('play_basic_beat2')
        self.advance_real_time(0.1)
        status = track_loops.get_status()
        self.assertEqual(2, len(status))
        self.assertEqual('delayed', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['kick2'].id, status[0]['sound_id'])
        self.assertEqual('playing', status[1]['status'])
        self.assertEqual(325660, status[1]['length'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[1]['sound_id'])
        self.assertEqual(status[0]['start_delay_samples_remaining'], status[1]['stop_loop_samples_remaining'])

        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='hi_hat_stopped')
        self.mc.bcp_processor.send.assert_any_call('trigger', name='basic_beat2_played')
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['kick2'].id, status[0]['sound_id'])

        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='basic_beat2_looping')
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['kick2'].id, status[0]['sound_id'])

    def test_sound_loop_track_layers(self):
        """ Tests the Sound Loop track layers"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudioSoundLoop')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudioSoundLoop')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Check sound loop track
        track_loops = interface.get_track_by_name("loops")
        self.assertIsNotNone(track_loops)
        self.assertEqual(track_loops.name, "loops")
        self.assertAlmostEqual(track_loops.volume, 0.6, 1)

        # /sounds/loops
        self.assertTrue(hasattr(self.mc, 'sounds'))
        self.assertIn('kick', self.mc.sounds)
        self.assertIn('kick2', self.mc.sounds)
        self.assertIn('hihat', self.mc.sounds)
        self.assertIn('snare', self.mc.sounds)
        self.assertIn('clap', self.mc.sounds)
        self.assertIn('bass_synth', self.mc.sounds)

        self.assertEqual(1, self.mc.sounds["kick"].marker_count)
        self.assertEqual(2, self.mc.sounds["hihat"].marker_count)

        # Sound loop sets
        self.assertTrue(hasattr(self.mc, 'sound_loop_sets'))
        self.assertIn('basic_beat_layers', self.mc.sound_loop_sets)
        self.assertIn('basic_beat_layers2', self.mc.sound_loop_sets)

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        # Test sound_loop_player
        self.advance_time()
        self.mc.events.post('play_sound_synthping')
        self.mc.events.post('play_basic_beat_layers')
        self.advance_real_time(0.1)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(325660, status[0]['length'])
        self.assertEqual(self.mc.sounds['kick'].id, status[0]['sound_id'])
        self.assertEqual(130.0, status[0]['tempo'])
        self.assertEqual(3, len(status[0]['layers']))
        self.assertEqual('stopped', status[0]['layers'][0]['status'])
        self.assertEqual('stopped', status[0]['layers'][1]['status'])
        self.assertEqual('stopped', status[0]['layers'][2]['status'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['layers'][0]['sound_id'])
        self.assertEqual(self.mc.sounds['snare'].id, status[0]['layers'][1]['sound_id'])
        self.assertEqual(self.mc.sounds['clap'].id, status[0]['layers'][2]['sound_id'])

        self.advance_real_time(0.9)

        # Ensure sound_loop_set.events_when_played is working properly (send event when a sound_loop_set is played)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='basic_beat_layers_played')

        self.mc.events.post('add_hi_hats')
        self.advance_time()
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(self.mc.sounds['kick'].id, status[0]['sound_id'])
        self.assertEqual(3, len(status[0]['layers']))
        self.assertEqual('queued', status[0]['layers'][0]['status'])
        self.assertEqual('stopped', status[0]['layers'][1]['status'])
        self.assertEqual('stopped', status[0]['layers'][2]['status'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['layers'][0]['sound_id'])
        self.assertEqual(self.mc.sounds['snare'].id, status[0]['layers'][1]['sound_id'])
        self.assertEqual(self.mc.sounds['clap'].id, status[0]['layers'][2]['sound_id'])

        self.advance_real_time(3)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(self.mc.sounds['kick'].id, status[0]['sound_id'])
        self.assertEqual(3, len(status[0]['layers']))
        self.assertEqual('playing', status[0]['layers'][0]['status'])
        self.assertEqual('stopped', status[0]['layers'][1]['status'])
        self.assertEqual('stopped', status[0]['layers'][2]['status'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['layers'][0]['sound_id'])
        self.assertEqual(self.mc.sounds['snare'].id, status[0]['layers'][1]['sound_id'])
        self.assertEqual(self.mc.sounds['clap'].id, status[0]['layers'][2]['sound_id'])

        self.mc.events.post('add_snare')
        self.mc.events.post('add_claps')
        self.advance_real_time(2)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(self.mc.sounds['kick'].id, status[0]['sound_id'])
        self.assertEqual(3, len(status[0]['layers']))
        self.assertEqual('playing', status[0]['layers'][0]['status'])
        self.assertEqual('playing', status[0]['layers'][1]['status'])
        self.assertEqual('playing', status[0]['layers'][2]['status'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['layers'][0]['sound_id'])
        self.assertEqual(self.mc.sounds['snare'].id, status[0]['layers'][1]['sound_id'])
        self.assertEqual(self.mc.sounds['clap'].id, status[0]['layers'][2]['sound_id'])

        # Ensure sound_loop_set.events_when_looping is working properly (send event when a sound_loop_set loops)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='basic_beat_layers_looping')

        # Ensure sound marker events are working properly for underlying sounds
        self.mc.bcp_processor.send.assert_any_call('trigger', name='kick_marker_1', sound_instance=ANY, marker_id=0)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='hihat_marker_1', sound_instance=ANY, marker_id=0)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='hihat_marker_2', sound_instance=ANY, marker_id=1)

        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(self.mc.sounds["kick"].id, status[0]['sound_id'])
        self.assertEqual(3, len(status[0]['layers']))
        self.assertEqual('playing', status[0]['layers'][0]['status'])
        self.assertEqual('playing', status[0]['layers'][1]['status'])
        self.assertEqual('playing', status[0]['layers'][2]['status'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['layers'][0]['sound_id'])
        self.assertEqual(self.mc.sounds['snare'].id, status[0]['layers'][1]['sound_id'])
        self.assertEqual(self.mc.sounds['clap'].id, status[0]['layers'][2]['sound_id'])
        self.advance_real_time(2)

        self.mc.events.post('play_basic_beat_layers2')
        self.advance_time()
        status = track_loops.get_status()
        self.assertEqual(2, len(status))
        self.assertEqual('delayed', status[0]['status'])
        self.assertEqual(self.mc.sounds["kick2"].id, status[0]['sound_id'])
        self.assertEqual('playing', status[1]['status'])
        self.assertEqual(self.mc.sounds["kick"].id, status[1]['sound_id'])
        self.assertEqual(status[0]['start_delay_samples_remaining'], status[1]['stop_loop_samples_remaining'])

        self.advance_real_time(4)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(self.mc.sounds["kick2"].id, status[0]['sound_id'])
        self.assertEqual(4, len(status[0]['layers']))
        self.assertEqual('playing', status[0]['layers'][0]['status'])
        self.assertEqual('playing', status[0]['layers'][1]['status'])
        self.assertEqual('stopped', status[0]['layers'][2]['status'])
        self.assertEqual('playing', status[0]['layers'][3]['status'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['layers'][0]['sound_id'])
        self.assertEqual(self.mc.sounds['snare'].id, status[0]['layers'][1]['sound_id'])
        self.assertEqual(self.mc.sounds['clap'].id, status[0]['layers'][2]['sound_id'])
        self.assertEqual(self.mc.sounds['bass_synth'].id, status[0]['layers'][3]['sound_id'])

        # Ensure sound_loop_set.events_when_stopped is working properly (send event when a sound_loop_set stops)
        self.mc.bcp_processor.send.assert_any_call('trigger', name='basic_beat_layers_stopped')
        self.mc.bcp_processor.send.assert_any_call('trigger', name='sound_loop_set_stopped')

        self.mc.events.post('fade_out_bass_synth')
        self.advance_time()
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(self.mc.sounds["kick2"].id, status[0]['sound_id'])
        self.assertEqual(4, len(status[0]['layers']))
        self.assertEqual('playing', status[0]['layers'][0]['status'])
        self.assertEqual('playing', status[0]['layers'][1]['status'])
        self.assertEqual('stopped', status[0]['layers'][2]['status'])
        self.assertEqual('fading out', status[0]['layers'][3]['status'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['layers'][0]['sound_id'])
        self.assertEqual(self.mc.sounds['snare'].id, status[0]['layers'][1]['sound_id'])
        self.assertEqual(self.mc.sounds['clap'].id, status[0]['layers'][2]['sound_id'])
        self.assertEqual(self.mc.sounds['bass_synth'].id, status[0]['layers'][3]['sound_id'])
        self.assertGreater(status[0]['layers'][3]['fade_out_steps'], 0)
        self.assertGreater(status[0]['layers'][3]['fade_steps_remaining'], 0)

        self.advance_real_time(4)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(self.mc.sounds["kick2"].id, status[0]['sound_id'])
        self.assertEqual(4, len(status[0]['layers']))
        self.assertEqual('playing', status[0]['layers'][0]['status'])
        self.assertEqual('playing', status[0]['layers'][1]['status'])
        self.assertEqual('stopped', status[0]['layers'][2]['status'])
        self.assertEqual('stopped', status[0]['layers'][3]['status'])
        self.assertEqual(self.mc.sounds['hihat'].id, status[0]['layers'][0]['sound_id'])
        self.assertEqual(self.mc.sounds['snare'].id, status[0]['layers'][1]['sound_id'])
        self.assertEqual(self.mc.sounds['clap'].id, status[0]['layers'][2]['sound_id'])
        self.assertEqual(self.mc.sounds['bass_synth'].id, status[0]['layers'][3]['sound_id'])
        self.assertGreater(status[0]['layers'][3]['fade_out_steps'], 0)
        self.assertEqual(0, status[0]['layers'][3]['fade_steps_remaining'])

        self.mc.events.post('reset_current_loop')
        self.advance_real_time(0.1)
        self.mc.events.post('reset_current_loop')
        self.advance_real_time(0.1)
        self.mc.events.post('reset_current_loop')
        self.advance_real_time(0.1)
        self.mc.events.post('reset_current_loop')
        self.advance_real_time(0.2)
        self.mc.events.post('reset_current_loop')
        self.advance_real_time(0.1)

        self.mc.events.post('play_basic_beat_layers')
        self.mc.events.post('stop_current_loop')
        self.mc.events.post('play_sound_synthping')
        self.advance_real_time(2)

        # Make sure next pending sound_loop_set is cancelled with stop action
        status = track_loops.get_status()
        self.assertEqual(0, len(status))

    def test_sound_loop_fades(self):
        """ Tests Sound Loop fading"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudioSoundLoop')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudioSoundLoop')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Check sound loop track
        track_loops = interface.get_track_by_name("loops")
        self.assertIsNotNone(track_loops)
        self.assertEqual(track_loops.name, "loops")
        self.assertAlmostEqual(track_loops.volume, 0.6, 1)

        # /sounds/loops
        self.assertTrue(hasattr(self.mc, 'sounds'))
        self.assertIn('hihat', self.mc.sounds)
        self.assertIn('kick', self.mc.sounds)
        self.assertIn('kick2', self.mc.sounds)

        # Sound loop sets
        self.assertTrue(hasattr(self.mc, 'sound_loop_sets'))
        self.assertIn('hi_hat', self.mc.sound_loop_sets)
        self.assertIn('basic_beat', self.mc.sound_loop_sets)
        self.assertIn('basic_beat2', self.mc.sound_loop_sets)

        # Play hi-hat loop and check status
        track_loops.play_sound_loop_set(self.mc.sound_loop_sets['hi_hat'])
        self.advance_real_time(0.1)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(status[0]['sound_id'], self.mc.sounds['hihat'].id)

        # Now play kick loop and recheck status (both loops should be cross-fading and in sync)
        track_loops.play_sound_loop_set(self.mc.sound_loop_sets['basic_beat'], None,
                                        {'fade_in': 1.0, 'timing': 'now', 'synchronize': True})
        self.advance_real_time(0.1)

        status = track_loops.get_status()
        self.assertEqual('fading out', status[1]['status'])
        self.assertEqual('fading in', status[0]['status'])
        self.assertGreater(status[1]['fade_out_steps'], 0)
        self.assertGreater(status[0]['fade_in_steps'], 1)
        self.assertEqual(status[1]['sample_pos'], status[0]['sample_pos'])
        self.assertEqual(status[1]['sound_id'], self.mc.sounds['hihat'].id)
        self.assertEqual(status[0]['sound_id'], self.mc.sounds['kick'].id)

        # Recheck status (hi-hat loop should be finished and kick loop should be playing)
        self.advance_real_time(1.1)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(status[0]['sound_id'], self.mc.sounds['kick'].id)

        # Now play hi hat loop and recheck status (both loops should be cross-fading and in sync)
        track_loops.play_sound_loop_set(self.mc.sound_loop_sets['hi_hat'], None,
                                        {'fade_in': 2.0, 'timing': 'now', 'synchronize': True})
        self.advance_real_time(0.1)

        status = track_loops.get_status()
        self.assertEqual(2, len(status))
        self.assertEqual('fading out', status[1]['status'])
        self.assertEqual('fading in', status[0]['status'])
        self.assertGreater(status[1]['fade_out_steps'], 0)
        self.assertGreater(status[0]['fade_in_steps'], 0)
        self.assertEqual(status[1]['sample_pos'], status[0]['sample_pos'])
        self.assertEqual(status[1]['sound_id'], self.mc.sounds['kick'].id)
        self.assertEqual(status[0]['sound_id'], self.mc.sounds['hihat'].id)

        self.advance_real_time(0.3)

        # Now play kick 2 and recheck status (new loop should be fading in and other two loops fading out)
        track_loops.play_sound_loop_set(self.mc.sound_loop_sets['basic_beat2'], None,
                                        {'fade_in': 0.8, 'timing': 'now', 'synchronize': False})
        self.advance_real_time(0.1)

        status = track_loops.get_status()
        self.assertEqual(3, len(status))
        self.assertEqual('fading in', status[0]['status'])
        self.assertEqual('fading out', status[1]['status'])
        self.assertEqual('fading out', status[2]['status'])
        self.assertGreater(status[0]['fade_in_steps'], 0)
        self.assertGreater(status[1]['fade_out_steps'], 0)
        self.assertGreater(status[1]['fade_out_steps'], 0)
        self.assertEqual(status[0]['fade_steps_remaining'], status[1]['fade_steps_remaining'])
        self.assertEqual(status[0]['fade_steps_remaining'], status[2]['fade_steps_remaining'])
        self.assertEqual(status[1]['sample_pos'], status[2]['sample_pos'])
        self.assertEqual(status[0]['sound_id'], self.mc.sounds['kick2'].id)
        self.assertEqual(status[1]['sound_id'], self.mc.sounds['hihat'].id)
        self.assertEqual(status[2]['sound_id'], self.mc.sounds['kick'].id)

    def test_sound_loop_timing_settings(self):
        """ Tests Sound Loop fading"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudioSoundLoop')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudioSoundLoop')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Check sound loop track
        track_loops = interface.get_track_by_name("loops")
        self.assertIsNotNone(track_loops)
        self.assertEqual(track_loops.name, "loops")
        self.assertAlmostEqual(track_loops.volume, 0.6, 1)

        # /sounds/loops
        self.assertTrue(hasattr(self.mc, 'sounds'))
        self.assertIn('hihat', self.mc.sounds)
        self.assertIn('kick', self.mc.sounds)
        self.assertIn('kick2', self.mc.sounds)

        # Sound loop sets
        self.assertTrue(hasattr(self.mc, 'sound_loop_sets'))
        self.assertIn('hi_hat', self.mc.sound_loop_sets)
        self.assertIn('basic_beat', self.mc.sound_loop_sets)
        self.assertIn('basic_beat2', self.mc.sound_loop_sets)

        # Play hi-hat loop and check status
        track_loops.play_sound_loop_set(self.mc.sound_loop_sets['hi_hat'])
        self.advance_real_time(0.1)
        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(status[0]['sound_id'], self.mc.sounds['hihat'].id)
        self.advance_real_time(0.5)

        # Now play kick loop and recheck status (loops should perform a quick cross-fade and switch)
        track_loops.play_sound_loop_set(self.mc.sound_loop_sets['basic_beat'], None, {'timing': 'now'})
        self.advance_real_time(0.2)

        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(status[0]['sound_id'], self.mc.sounds['kick'].id)
        self.advance_real_time(0.3)

        # Now play second kick loop and recheck status (loops should perform a quick cross-fade and switch)
        track_loops.play_sound_loop_set(self.mc.sound_loop_sets['basic_beat2'], None, {'timing': 'now'})
        self.advance_real_time(0.2)

        status = track_loops.get_status()
        self.assertEqual(1, len(status))
        self.assertEqual('playing', status[0]['status'])
        self.assertEqual(status[0]['sound_id'], self.mc.sounds['kick2'].id)
        self.advance_real_time(0.3)
